# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import data_splitting
import complexity_weights
import math_block
import math_block_set
from model_evaluation import ModelEvaluation
from solution import Solution
import time
import variable_options
from variable_options_dict import VariableOptionsDict
import json
from utils import Throttle
import warnings

class Search:
    """Represents a search on the server. It should be created using the
    :py:meth:`~eureqa.data_source.DataSource.create_search` method. A search can be run
    using the :py:meth:`~eureqa.search.Search.submit` method.
    Once a search has ran, solutions can be retreived from one of several methods below that return
    a :py:class:`~eureqa.solution.Solution`, including :py:meth:`~eureqa.search.Search.get_best_solution`
    to get the best solution and :py:meth:`~eureqa.search.Search.get_solutions` to get a list of all solutions.

    Many of the variables will be set based on the :py:class:`~eureqa.search_settings.SearchSettings` used when
    :py:meth:`~eureqa.data_source.DataSource.create_search` is called. However, these variables can still be changed
    after the Search's creation.

    :var str `~eureqa.search.Search.name`: The name of the search.
    :var MathBlockSet `~eureqa.search.Search.math_blocks`:
        The MathBlockSet which represents mathematical operations allowed to be
        used in the search's solutions. See the usage section of :py:class:`~eureqa.search_settings.SearchSettings` to see
        examples of how to change it.
    :var DataSplitting data_splitting: A DataSplitting object that holds data splitting settings for the search algorithm.
    :var str error_metric: The error metric that will be used for this search. Choose from the error metrics in :mod:`~eureqa.error_metric`.
    :var int maximum_history_absolute_rows: The maximum number of rows that can be used in range based functions.
    :var list prior_solutions: The list of prior solutions.
    :var str row_weight: The row weight expression.
    :var str target_expression: The target expression.
    :var VariableOptionsDict variable_options: Override default behavior for the specified variables.
    """

    def __init__(self, body, eureqa, datasource=None):
        """For internal use only
        PARAM_NOT_EXTERNALLY_DOCUMENTED
        """

        self._id = body['search_id']
        self.name = body['search_name']
        self.math_blocks = math_block_set.MathBlockSet.from_json(body['building_blocks'])
        self._complexity_weights = complexity_weights._from_json(body['complexity_weights'])
        self.data_splitting = data_splitting._from_json(body)
        self._data_source_id = body['datasource_id']
        self.error_metric = body['error_metric']
        self._max_num_variables_per_term = body['max_num_variables_per_term']
        self.maximum_history_absolute_rows = body.get('max_delay')
        self.prior_solutions = body.get('prior_solutions', [])
        self.row_weight = body['row_weight']
        self.row_weight_type = body['row_weight_type']
        self.target_expression = body['target_expression']
        self.default_min_delay = body.get('default_min_delay')
        self.variable_options = VariableOptionsDict.from_json(body['variable_options'])
        self._eureqa = eureqa
        self._body = body
        self._datasource_object = datasource

    def _to_json(self):
        data_splitting_settings = {}
        self.data_splitting._to_json(data_splitting_settings)
        body_updates = {
            'name': self.name,
            'math_blocks': self.math_blocks._to_json(),
            'data_splitting': data_splitting_settings,
            'error_metric': self.error_metric,
            'maximum_history_absolute_rows': self.maximum_history_absolute_rows,
            'prior_solutions': self.prior_solutions,
            'row_weight': self.row_weight,
            'target_expression': self.target_expression,
            'variable_options': self.variable_options._to_json()
        }
        body = dict(self._body)
        body.update(body_updates)
        return body

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    @property
    def math_blocks(self):
        """ The MathBlockSet which represents mathematical operations allowed in solutions to the Search.

        """
        return self._math_blocks
    @math_blocks.setter
    def math_blocks(self,val):
        self._math_blocks = val

    def get_data_source(self):
        """ Retrieves from the server the data source information for this search.

        :rtype: DataSource
        """

        if not self._datasource_object:
            self._datasource_object = self._eureqa.get_data_source_by_id(self._data_source_id)
        return self._datasource_object


    def delete(self):
        """Deletes the search from the server.

        :raise Exception: search is already deleted.
        """

        endpoint = '/fxp/datasources/%s/searches/%s' % (self._data_source_id, self._id)
        self._eureqa._session.report_progress('Deleting search: \'%s\' from datasource: \'%s\'.' % (self._id, self._data_source_id))
        self._eureqa._session.execute(endpoint, 'DELETE')

    def submit(self, time_seconds):
        """Submit the search to the server to run for the specified amount of time.

        This method does not guarantee to start the search immediately. The search can
        be queued for some time before it will start producing any results.

        :param int time_seconds:
            The maximum amount of time to run the search. The server will stop
            running the search once the running time will reach this limit.
        """


        endpoint = '/fxp/search_queue'
        body = {'datasource_id': self._data_source_id,
                'search_id': self._id,
                'stopping_condition':
                    {'time_seconds': time_seconds,
                     'max_r2': 1}}
        self._eureqa._session.report_progress('Submitting search: \'%s\' from datasource: \'%s\' to the search queue.' % (self._id, self._data_source_id))
        self._eureqa._session.execute(endpoint, 'POST', body)

    @property
    def is_running(self):
        """Indicates if the search currently running.

        :rtype: bool
        """
        search_status = self._get_updated()._body['search_queue_state']
        # The endpoint is returning the internal state of the search queue, which treats
        # queueing states as not running.
        return search_status not in {'', 'QS_NONE', 'QS_DONE', 'QS_ABORTED'}

    def stop(self):
        """Stops running the search."""

        endpoint = '/fxp/search_queue/stop'
        body = {'datasource_id': self._data_source_id, 'search_id': self._id}
        self._eureqa._session.report_progress('Stopping search: \'%s\' from datasource: \'%s\'.' % (self._id, self._data_source_id))
        self._eureqa._session.execute(endpoint, 'POST', body)
        self.wait_until_done()

    def wait_until_done(self, show_progress=False, poll_seconds=5, print_callback=None):
        """Waits until the search stops running.

        :param bool show_progress: whether to print the search progress while waiting.
        :param int poll_seconds: number of seconds to wait between checking progress.
        :param function print_callback: method to invoke to print the progress (sys.stdout.write by default).
        """

        if print_callback is None:
            import sys
            print_callback = sys.stdout.write

        prev_solution_string = ''
        while self.is_running:

            if show_progress:
                progress_string = self._get_progress_summary()
                solution_string = self._get_solutions_summary()
                print_callback(progress_string)
                if solution_string != prev_solution_string:
                    prev_solution_string = solution_string
                    print_callback('\n%s' % solution_string)

            time.sleep(poll_seconds)

    def _get_progress_summary(self):
        """Returns convenience values of the search progress."""
        search = self._get_updated()._body
        return 'Search="%s", remaining: %is, converged=%.2g%%, cores=%i, evals=%.4gM (%.2gM e/s), gens=%g (%.4g g/s)' % (
            search['search_name'],
            search['estimated_time_remaining_seconds'],
            search['search_stats']['search_percent_converged'],
            search['search_stats']['search_cpu_cores'],
            search['search_stats']['search_evaluations']/1e6, search['search_stats']['search_evaluations_per_second']/1e6,
            search['search_stats']['search_generations'],     search['search_stats']['search_generations_per_second'])

    def _get_solutions_summary(self):
        ret = ''
        for s in self.get_solutions():
            ret += '%i, %.4g, [ %s ]\n' % (s.complexity,
                s.optimized_error_metric_value, ', '.join(s.terms))
        return ret

    @Throttle()
    def get_solutions(self):
        """Retrieves from the server the list of solutions found so far.

        This method can be called while the search is running to check what searches are
        already found and make a decision whether continue the search.

        :rtype: list of :py:class:`~eureqa.solution.Solution` objects.
        """

        endpoint = '/fxp/datasources/%s/searches/%s/solutions' % (self._data_source_id, self._id)
        self._eureqa._session.report_progress('Getting solutions for search: \'%s\' from datasource: \'%s\'.' % (self._id, self._data_source_id))
        body = self._eureqa._session.execute(endpoint, 'GET')
        return [Solution(x, self._eureqa, self) for x in body]

    def get_best_solution(self):
        """Retrieves from the server the best solution found so far.

        :rtype: Solution
        """

        solutions = self.get_solutions()
        return next((x for x in solutions if x.is_best), None)

    def get_most_accurate_solution(self):
        """Retrieves from the server the most accurate solution found so far.

        :rtype: Solution
        """

        solutions = self.get_solutions()
        if not solutions: return None
        return next((x for x in solutions if x.is_most_accurate), None)

    def create_solution(self, solution_string, use_all_data=False):
        """Creates a custom solution for the search.
        Use this if you want to compute error metrics and other statistics of a specified expression.
        It is also useful to compare how well a known model does against one found by Eureqa

        :param string solution_string: the right hand side of the expression.
        :param bool use_all_data: whether to use all data or just validation data when calculating the metrics for the solution.

        :rtype: Solution
        """

        endpoint = '/fxp/datasources/%s/searches/%s/solutions' % (self._data_source_id, self._id)
        body = {'datasource_id': self._data_source_id,
                'search_id': self._id,
                'solution_string': solution_string,
                'use_all_data': use_all_data}
        result = self._eureqa._session.execute(endpoint, 'POST', body)
        return Solution(result, self._eureqa, self)

    @Throttle()
    def _get_updated(self):
        return self._eureqa._get_search_by_search_id(self._data_source_id, self._id)

    def evaluate_expression(self, expressions, _data_split='all'):
        """Deprecated. Use :py:meth:`~eureqa.eureqa.Eureqa.evaluate_expression()` instead

        :param str expressions: Deprecated.  Do not use.
        """
        warnings.warn("This function has been deprecated.  Please use `Eureqa.evaluate_expression()` instead.", DeprecationWarning)
        return self._eureqa.evaluate_expression(self._data_source_id, expressions, self, _data_split=_data_split)

    def rename(self, new_search_name):
        """Change search display name.

        :param std new_search_name: New search name.
        """

        endpoint = '/fxp/datasources/%s/searches/%s/rename' % (self._data_source_id, self._id)
        body = {'new_search_name': new_search_name}
        self._eureqa._session.execute(endpoint, 'POST', body)
        self.name = new_search_name

    def _evaluate_solutions(self, include_data=False, calculate_error_metrics=False):
        include_data_string = 'true' if include_data else 'false'
        calculate_error_metrics = 'true' if calculate_error_metrics else 'false'
        endpoint = '/fxp/datasources/%s/searches/%s/evaluate_models?include_data=%s&calculate_error_metrics=%s' % (
                self._data_source_id, self._id, include_data_string, calculate_error_metrics)
        response = self._eureqa._session.execute(endpoint, 'GET')
        return ModelEvaluation(response)







# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import csv
from string import ascii_uppercase, digits
from StringIO import StringIO
from random import choice
from collections import OrderedDict

class DataHolder:
    """
    For internal use only.

    A helper class for holding arbitrary data in a 2d list, then writing to a CSV in a StringIO buffer.
    """

    delimiter = ','

    def __init__(self):
        self._cols = OrderedDict()

    def add_column(self, name, column):
        """Add a column of data with the specified name to the 2d array.

        :param str name: The name of the column.
        :param list column: The column of data.

        If the name is already used in this instance, the old column will be overwritten.
        """
        if not isinstance(column, list):
            raise Exception("Error in DataHolder.add_column: all input values must be lists, not \"" + str(type(column)) + "\"")

        #if self._cols.has_key(name):
        #    print "Warning in DataHolder.add_column: variable \"" + name + "\" already exists"

        self._cols[name] = column
        self._pad_with_nans()

    def _largest_col_length(self):
        """Return the length of the largest column"""
        return max([0] + [len(self._cols[name]) for name in self._cols])

    def _pad_with_nans(self):
        """Fill all columns with None so they're the same length as the largest column"""
        for name in self._cols.keys():
            self._cols[name] += [None]*(self._largest_col_length() - len(self._cols[name]))

    def get_csv_file(self):
        """
        Translate the accumulated 2d array of data into a CSV file inside a StringIO buffer.

        :return: a StringIO buffer containing the CSV
        :rtype: StringIO
        """
        buf = StringIO()
        csv_writer = csv.writer(buf, delimiter=self.delimiter)
        names = self._cols.keys()
        csv_writer.writerow(names)

        for i in xrange(self._largest_col_length()):
            row = [self._cols[name][i] for name in names]
            csv_writer.writerow(row)
        buf.seek(0)
        # Our API requires a file name ending in csv to accept the file. Attaching this to
        # the file-like object allows the request module to handle this
        buf.name = ''.join(choice(ascii_uppercase + digits) for i in range(10)) + '.csv'
        return buf



from eureqa.analysis.components.base import _Component

class Tooltip(_Component):
    """
    Tooltip Component: implements a tooltip

    For example::

        tt = Tooltip(html="This", tooltip="Text to show when hovering")
        card = analysis.create_html_card("This is a component with a tooltip: {0}".format(analysis.html_ref(tt)))


    :param str html: Text to display normally
    :param str tooltip: The tooltip text to show when the cursor hovers over the component
    """

    _component_type_str = 'TOOLTIP'

    def __init__(self, html=None, tooltip=None, _analysis=None, _component_id=None, _component_type=None):

        super(Tooltip, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)

        if html is not None:
            self._text = html

        if tooltip is not None:
            self._tooltip_content = tooltip


    @property
    def html(self):
        """The text to show normally for this component.

        :rtype: str
        """
        return self._text

    @html.setter
    def html(self, val):
        self._text = val
        self._update()


    @property
    def tooltip(self):
        """The tooltip text to show when the cursor hovers over the component

        :rtype: str
        """
        return self._tooltip_content

    @tooltip.setter
    def tooltip(self, val):
        self._tooltip_content = val
        self._update()



    def _fields(self):
        return super(Tooltip, self)._fields() + [ 'text', 'tooltip_content' ]



# This value has to equal to the version of the REST API.
# It should be incremented every time REST API is changed in the way that
# can break this API binding.
# For example, if an end-point used by this API binding is changed.
# Or if a new endpoint is added and used by a new version of this API binding.

API_VERSION = 13



from eureqa.analysis.components.base import _Component
from eureqa.utils.jsonrest import _JsonREST
from eureqa.analysis.analysis_file import AnalysisFile

class TabbedLayout(_Component):
    """
    TabbedLayout component

    Lets users pick to see other components based on titles (works with a few items). Similar to :py:class:`~eureqa.analysis.components.dropdown_layout.DropdownLayout` but suited to smaller lists

    For example::

        t = TabbedLayout()
        t.add_component(title="choice 1", component=HtmlBlock("This is item 1"))
        t.add_component(title="choice 2", component=HtmlBlock("This is item 2"))
        analysis.create_card(t)

    """

    _component_type_str = 'TABBED_LAYOUT'

    def __init__(self, _tabs=None, _analysis=None, _component_id=None, _component_type=None):
        # Tabs to add as soon as we are _register()'d
        # Note that `_queued_tabs` are _Tab objects and `_tabs` are JSON dictionaries
        self._queued_tabs = []

        if _tabs is not None:
            self._queued_tabs = _tabs

        super(TabbedLayout, self).__init__(_analysis=_analysis, _component_id=_component_id,
                                           _component_type=_component_type)

    class _Tab(object):
        def __init__(self, title, component, icon=None):
            """
            :param str label: Tab label
            :param ~eureqa.analysis.components.base._Component component: Tab contents, as a Component
            :param ~eureqa.analysis.analysis_file.AnalysisFile label_img: Tab label image (optional) -- must be a file containing an image
            """
            self._title = title
            self._component = component
            self._icon = icon

        def _register(self, analysis):
            if self._icon is not None and not isinstance(self._icon, AnalysisFile):
                self._icon = AnalysisFile.create(analysis, self._icon)

        def __repr__(self):
            return repr([self._title, self._component, str(self._icon)[:32] if self._icon else self._icon])

        def _to_json(self):
            ret = {
                "label": self._title,
                "icon_file_id": self._icon._file_id if getattr(self, "_icon", None) is not None else None
            }
            if hasattr(self._component, "_component_id"):
                ret["component_id"] = self._component._component_id
            return ret

    def _ensure_tabs(self):
        if not hasattr(self, "_tabs"):
            self._tabs = []
        if not hasattr(self, "_tab_components"):
            self._tab_components = []
            for tab_json in self._tabs:
                tab = TabbedLayout._Tab(tab_json["label"], _Component._get(_analysis=self._analysis,
                                                            _component_id=tab_json["component_id"]))
                tab._icon = AnalysisFile(self._analysis, tab_json["icon_file_id"])
                self._tab_components.append(tab)

    def add_component(self, title, component, icon=None):
        """
        Add a new tab containing a new component

        :param str title: Title of the tab
        :param eureqa.analysis.components.base._Component component: Content of the tab
        :param str icon: (Optional) Icon for the tab, as either a str() containing the file-data for an image, or an :class:`AnalysisFile` containing an image
        """
        new_tab = TabbedLayout._Tab(title, component, icon)

        if hasattr(self, "_analysis") and hasattr(new_tab, "_component_id"):
            new_tab._associate(self._analysis)
            self._ensure_tabs()
            self._tabs.append(new_tab._to_json())
            self._tab_components.append(new_tab)
        else:
            self._queued_tabs.append(new_tab)

    def _walk_children(self):
        for c in getattr(self, "_tab_components", []):
            yield c._component
        for c in self._queued_tabs:
            yield c._component

    def _register(self, analysis):
        super(TabbedLayout, self)._register(analysis)
        self._ensure_tabs()
        for new_tab in self._queued_tabs:
            new_tab._register(analysis)
            self._tabs.append(new_tab._to_json())
            self._tab_components.append(new_tab)
        self._queued_tabs = []

    def _fields(self):
        return super(TabbedLayout, self)._fields() + [ 'tabs' ]



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from analysis import Analysis
from complexity_weights import ComplexityWeights
from data_source import DataSource
import data_splitting
from data_splitting import DataSplitting
import error_metric
from eureqa import Eureqa
import math_block
from math_block import MathBlock
import missing_value_policies
from search import Search
import search_settings
from search_settings import SearchSettings
from solution import Solution
import variable_options
from variable_options import VariableOptions

__all__ = ['Analysis', 'ComplexityWeights', 'DataSource', 'data_splitting',
           'DataSplitting', 'error_metric', 'Eureqa', 'math_block', 'MathBlock', 'missing_value_policies',
           'Search', 'search_settings', 'SearchSettings', 'Solution', 'variable_options', 'VariableOptions']



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os.path

from eureqa.analysis.analysis_file import AnalysisFile
from eureqa.analysis.components.modal_link import ModalLink
from eureqa.analysis.components.base import _Component

class Modal(_Component):
    """Represents a popup window (modal) which can contain other components.

    Example:
    # create a modal which is activated if a link is clicked
    text_block = TextBlock("This text is inside the modal")
    modal = Modal(title="This is the modal title", size='medium', icon_file_path='path/to/icon', component=text_block)
    h = HtmlBlock('Click this link to open a modal: {0}'.format(analysis.html_ref(modal.link('link text'))))
    layout.add_component(h)

    :param str title: The title of the modal
    :param str size: (optional) modal size. Options are 'small', 'medium' or 'large'
    :param str icon_file_path: (optional) path to an icon file to display in the corner of the modal
    :param Component component: The component to display in the modal
    """

    _component_type_str = 'MODAL'

    def __init__(self, title=None, size=None, icon_file_path=None, component=None,
                 _analysis=None, _component_id=None, _component_type=None):
        self._title = title or ''
        self._size = size or 'small'
        if icon_file_path is not None:
            self._icon_file_path = icon_file_path
        if component is not None:
            self._content_component = component

        super(Modal, self).__init__(_analysis=_analysis, _component_id=_component_id,
                                    _component_type=_component_type)

        # call to self._associate may have been triggered by super.__init__ above
        # so we set up the content component after that call, to get the component id if there is one
        if hasattr(self, '_content_component') and hasattr(component, '_component_id'):
                self._content_component_id = component._component_id

    @property
    def content_component_id(self):
        """The id of the component to be displayed in the modal

        :rtype: str
        """
        return getattr(self, "_content_component_id", None)

    @property
    def title(self):
        """The title of the modal

        :rtype: str
        """
        return getattr(self, "_title", None)

    @property
    def size(self):
        """The size of the modal

        :rtype: str
        """
        return getattr(self, "_size", None)

    @property
    def icon_file_url(self):
        """The url of an icon file to be displayed in the corner of the modal

        :rtype: str
        """
        return getattr(self, "_icon_file_url", None)

    @property
    def icon_file_id(self):
        """The id of an icon file to be displayed in the corner of the modal

        :rtype: str
        """
        return getattr(self, "_icon_file_id", None)

    def _walk_children(self):
        if hasattr(self, "_content_component"):
            yield self._content_component

    def _register(self, analysis):
        updated_content = False
        if hasattr(self, '_content_component'):
            self._content_component_id = self._content_component._component_id
            updated_content = True

        if hasattr(self, '_icon_file_path') and not hasattr(self, "_icon_file_url"):
            icon_file_name = os.path.basename(self._icon_file_path)
            icon_file = open(self._icon_file_path, 'rb')
            self._icon_file = AnalysisFile.create(analysis, icon_file, filename=icon_file_name)
            self._icon_file_url = ("/api/%s/analysis/%s/files/%s" % (analysis._organization, analysis._id, self._icon_file._file_id))
            self._icon_file_id = self._icon_file._file_id
            updated_content = True

        super(Modal, self)._register(analysis)

    def _associate_with_table(self, analysis):
        content_json = []
        if hasattr(self, '_content_component'):
            content_json += self._content_component._associate_with_table(analysis)
            self._content_component_id = self._content_component._component_id

        if hasattr(self, '_icon_file_path'):
            icon_file_name = os.path.basename(self._icon_file_path)
            icon_file = open(self._icon_file_path, 'rb')
            self._icon_file = AnalysisFile.create(analysis, icon_file, filename=icon_file_name)
            self._icon_file_url = ("/api/%s/analysis/%s/files/%s" % (analysis._organization, analysis._id, self._icon_file._file_id))
            self._icon_file_id = self._icon_file._file_id

        return super(Modal, self)._associate_with_table(analysis) + content_json

    def _fields(self):
        return super(Modal, self)._fields() + [ 'content_component_id', 'title', 'size', 'icon_file_url', 'icon_file_id' ]

    def link(self, link_text=None):
        """
        Returns a ModalLink component which represents a link to this modal

        :param str link_text: (optional) link text to use in the ModalLink instead of default text
        :rtype: eureqa.analysis.components.modal_link.ModalLink
        """
        return ModalLink(modal=self, link_text=link_text)




# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from error_metric import ErrorMetrics

import json

class Solution:
    """Represents one of the solutions found by the server for a particular search. Can be obtained from
    one of the methods in :py:class:`~eureqa.search.Search`. Some solutions may be available even if the
    search has not finished running.

    :var str `~eureqa.solution.Solution.target`: The target variable.
    :var str `~eureqa.solution.Solution.model`: The model expression.
    :var int `~eureqa.solution.Solution.complexity`: The model complexity based on the complexity weights.
    :var bool `~eureqa.solution.Solution.is_best`:
        An indicator whether the solution is considered to be the best based on its complexity and precision.
        The system can pick only one solution as best.
    :var bool `~eureqa.solution.Solution.is_most_accurate`:
        An indicator whether the solution is the most accurate for the optimized error metric.
        The system can pick only one solution as most accurate.
    :var str `~eureqa.solution.Solution.optimized_error_metric`:
        The error metric for which the solution was optimized.
        It is one of the error metrics from :mod:`~eureqa.error_metric`.
    :var float `~eureqa.solution.Solution.optimized_error_metric_value`: The value of the optimized error metric.
    :var ~eureqa.search.Search `~eureqa.solution.Solution.search`: The search object to which this solution belongs.
    """

    def __init__(self, body, eureqa, search=None):
        """For internal use only.
        PARAM_NOT_EXTERNALLY_DOCUMENTED
        """

        self._id = body['solution_id']
        self._search_id = body['search_id']
        self._datasource_id = body['datasource_id']
        self.target = body['target_variable']
        self.model = body['model_string']
        self.complexity = body['model_complexity']
        self.is_best = body['solution_type']['best']
        self.is_most_accurate = body['solution_type']['most_accurate']
        self._error_metrics = {x['metric_name']: x['metric_value'] for x in body['error_metric_values']}
        self.optimized_error_metric = body['model_optimized_error_metric']
        self.optimized_error_metric_value = body['model_optimized_error_metric_value']
        self._model_fitness = body['model_fitness']
        self._model_aic_parameters = body.get('model_aic_parameters')
        self.terms = [t['term_string'] for t in body['terms']]
        self._body = body
        self._eureqa = eureqa
        self._search_object = search

    def _to_json(self):
        body = {
            'target': self.target,
            'model': self.model,
            'complexity': self.complexity,
            'is_best': self.is_best,
            'is_most_accurate': self.is_most_accurate,
            'optimized_error_metric': self.optimized_error_metric,
            'optimized_error_metric_value': self.optimized_error_metric_value,
            'terms': self.terms
        }
        return body

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    def get_error_metric_value(self, name):
        """
        Returns the value for the specified error metric.

        :param str name: One of the error metrics from :mod:`~eureqa.error_metric`.
        :rtype: float
        """

        return self._error_metrics[name]

    def get_all_series_error_metrics(self, data_split='all'):
        """
        Returns the metric values for each series in the datasource.

        :param str data_split: For internal use only.

        :return: list of :class:`~eureqa.error_metric.ErrorMetrics` objects.
        :rtype: list
        """

        result = self.search._eureqa._session.execute('/fxp/datasources/%s/searches/%s/solutions/%s/series_metrics?data_type=%s'% (self.search._data_source_id, self.search._id, self._id, data_split), 'GET')
        metrics_list = []
        for x in result:
            metrics = ErrorMetrics()
            metrics._from_json(x)
            metrics_list.append(metrics)
        return metrics_list

    def get_single_series_error_metrics(self, series=None, data_split='all', series_index=None):
        """
        Returns the metric values for the specified series in the datasource.

        :param int,str series: The series to compute the metrics on. If series is an integer, returns the error metrics for the series with that series index. Otherwise, returns the error metrics for the series with the specified series id value.
        :param str data_split: For internal use only.
        :param int series_index: Deprecated. Only kept for backwards compatibility. Do not use.
        :rtype: ErrorMetrics
        """

        assert series_index is None and series is not None or\
               series_index is not None and series is None

        if series_index is not None:
            series=series_index

        if isinstance(series, (int, long)):
            result = self.search._eureqa._session.execute('/fxp/datasources/%s/searches/%s/solutions/%s/series_metrics/%s?data_type=%s'% (self.search._data_source_id, self.search._id, self._id, series, data_split), 'GET')
            metrics = ErrorMetrics()
            metrics._from_json(result)
            return metrics
        else:
            result = self.search._eureqa._session.execute('/fxp/datasources/%s/searches/%s/solutions/%s/series_metrics?data_type=%s&series_id_value=%s'% (self.search._data_source_id, self.search._id, self._id, data_split, series), 'GET')
            assert len(result) == 1
            metrics = ErrorMetrics()
            metrics._from_json(result[0])
            return metrics

    @property
    def search(self):
        if not self._search_object:
            self._search_object = self._eureqa._get_search_by_search_id(self._datasource_id, self._search_id)
        return self._search_object



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from eureqa.analysis.components.base import _TwoVariablePlotComponent
from eureqa.utils.jsonrest import _JsonREST

class DoubleHistogramPlot(_TwoVariablePlotComponent):
    """Represents a double-histogram plot component on the server.

    For example::

      p = DoubleHistogramPlot(d, axis_labels={'x': 'the x var', 'y' : 'the y var'}, label_format={'y':'.3s'}, x_var='W', y_var='A', needs_guides=True)
      analysis.create_card(p)


    :param DataSource datasource: The data source containing the data to be plotted.
    :param str x_var: The X-axis variable for the card's plot. (must be binary)
    :param str y_var: The Y-axis variable for the card's plot.
    :param bool needs_guides: Whether the card needs guides.
    :param dict axis_labels: Axis labels for this card's plot.  Set keys "x" and "y" to set the X and Y axis labels.
    :param dict label_format: Label format for this card.  Set keys "x" and "y" to set the X and Y axis printf-style format-strings; for example, ".3s".

    :var str ~eureqa.analysis_cards.two_variable_plot.TwoVariablePlot.title: The title of the card
    :var str ~eureqa.analysis_cards.two_variable_plot.TwoVariablePlot.x_var: The X-axis variable for the card's plot (must be binary)
    :var str ~eureqa.analysis_cards.two_variable_plot.TwoVariablePlot.y_var: The Y-axis variable for the card's plot
    :var bool ~eureqa.analysis_cards.two_variable_plot.TwoVariablePlot.needs_guides: Whether the card needs guides
    :var XYMap ~eureqa.analysis_cards.two_variable_plot.TwoVariablePlot.axis_labels: Axis labels for this card's plot.  Set member fields "x" and "y" to set the X and Y axis labels.
    :var XYMap ~eureqa.analysis_cards.two_variable_plot.TwoVariablePlot.label_format: Label format for this card.  Set member fields "x" and "y" to set the X and Y axis printf-style format-strings; for example, ".3s".
    """

    _component_type_str = 'DOUBLE_HISTOGRAM_PLOT'

    def __init__(self, datasource=None, axis_labels=None,
                 label_format=None, needs_guides=None, x_var=None, y_var=None,
                 _analysis=None, _component_id=None, _component_type=None):

        if datasource is not None:
            # Set the property.  Also sets the ID.
            self.datasource = datasource

        if axis_labels is not None:
            self.axis_labels = axis_labels

        if label_format is not None:
            self.label_format = label_format

        if needs_guides is not None:
            self.needs_guides = needs_guides

        if x_var is not None:
            self.x_var = x_var

        if y_var is not None:
            self.y_var = y_var

        super(DoubleHistogramPlot, self).__init__(_analysis=_analysis, _component_id=_component_id,
                                                  _component_type=_component_type)

    @property
    def datasource(self):
        """The data source providing data for this component

        :return: data source providing data for this card
        """
        if hasattr(self, "_eureqa"):
            return self._eureqa.get_data_source_by_id(self._datasource_id)

    @datasource.setter
    def datasource(self, val):
        self._datasource_id = val._data_source_id
        self._eureqa = val._eureqa
        self._update()

    @property
    def axis_labels(self):
        """The axis labels for this card

        defaults to:
        ::

            { 'x': x_var, 'y': y_var }

        :return: Axis labels for this card
        :rtype: XYMap
        """
        return self.XYMap(self, getattr(self, "_axisLabels", {"x": "", "y": ""}))

    @axis_labels.setter
    def axis_labels(self, val):
        if hasattr(val, "x") and hasattr(val, "y"):
            val = {"x": val.x, "y": val.y}
        self._axisLabels = val
        self._update()

    @property
    def label_format(self):
        """Label format for this card.  Set keys "x" and "y" to set the X and Y axis printf-style format-strings; for example, ".3s".

        defaults to:
        ::

            { 'x': '.3s', 'y': 'g' }

        :rtype: XYMap
        """
        return self.XYMap(self, getattr(self, "_labelFormat", {"x": "", "y": ""}))

    @label_format.setter
    def label_format(self, val):
        if hasattr(val, "x") and hasattr(val, "y"):
            val = {"x": val.x, "y": val.y}
        self._labelFormat = val
        self._update()

    @property
    def needs_guides(self):
        """Does this card need guides?

         :return: Whether this card needs guides
         :rtype: bool
         """
        return getattr(self, "_needsGuides", None)

    @needs_guides.setter
    def needs_guides(self, val):
        self._needsGuides = val
        self._update()

    @property
    def x_var(self):
        """The X variable for this card. (must be binary)

        :return: X variable for this card
        :rtype: str
        """
        return getattr(self, "_x_var", None)

    @x_var.setter
    def x_var(self, val):
        self._x_var = val
        self._update()

    @property
    def y_var(self):
        """The Y variable for this card.

        :return: Y variable for this card
        :rtype: str
        """
        return getattr(self, "_y_var", None)

    @y_var.setter
    def y_var(self, val):
        self._y_var = val
        self._update()

    def _fields(self):
        return super(DoubleHistogramPlot, self)._fields() + [ 'datasource_id', 'axisLabels', 'description', 'labelFormat', 'needsGuides', 'x_var', 'y_var' ]



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json

class MathBlock(object):
    """Represents a building block of a mathematical model. Created
    automatically during the construction of a :py:class:`~eureqa.search_settings.SearchSettings` object.

    MathBlocks can be enabled or disabled to specify whether their corresponding
    mathematical operations are allowed in a :py:class:`~eureqa.search.Search`'s solutions.
    They should be accessed through the :py:attr:`eureqa.search.Search.math_blocks` 
    or :py:attr:`eureqa.search_settings.SearchSettings.math_blocks` properties.
    For a list of available Math Blocks, view the properties of :py:class:`~eureqa.math_block_set.MathBlockSet`.

    """

    def __init__(self, id_, name, complexity, notation):
        """For internal use only.
        PARAM_NOT_EXTERNALLY_DOCUMENTED
        """

        
        self._name = name
        self._id = id_
        self._complexity = complexity
        self._default_complexity = complexity
        self._notation = notation
        self._enabled = False

    @property
    def name(self):
        """ MathBlock's name (read-only) """
        return self._name

    @property
    def complexity(self):
        """ MathBlock's complexity (settable) """
        return self._complexity

    @complexity.setter
    def complexity(self, val):
        self._complexity = val

    @property
    def enabled(self):
        """ Can this MathBlock be used for modeling? """
        return self._enabled

    def enable(self, complexity=None):
        """ Allow this MathBlock to be used for modeling

        :param int complexity: The level of complexity of the MathBlock. If you do not specify a complexity, the default complexity will be used.

        """
        if complexity is None:
            complexity = self._default_complexity
        assert complexity is not None, "Called 'enable()' without specifying a complexity on a MathBlock that's not fully constructed"

        self.complexity = complexity

        self._enabled = True
            
    def disable(self):
        """ Don't allow this MathBlock to be used for modeling """
        self._enabled = False

    def _from_json(self, body):
        self._name = body['op_name']
        self._complexity = int(body['complexity'])
        self._id = int(body['op_id'])
        # only some endpoints return the notation field.
        if 'op_notation' in body:
            self._notation = body['op_notation']
        self._enabled = body.get('enabled', True)
        self._body = body

    @classmethod
    def from_json(cls, body):
        mathBlock = MathBlock(None, None, None, None)
        mathBlock._from_json(body)
        return mathBlock

    def _to_json(self):
        body = {'op_name': self.name,
                'complexity': self.complexity,
                'op_id': self._id,
                'enabled': self._enabled}
        if self._notation is not None:
            body['op_notation'] = self._notation
        return body

    def __eq__(self, other):
        if not isinstance(other, MathBlock):
            return False
        return other._id == self._id and other.name == self.name and other.complexity == self.complexity

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        tpl = (self._id, self.name, self.complexity)
        return hash(tpl)

    def __str__(self):
        return json.dumps(self._to_json(), indent=4, sort_keys=True)

    def __repr__(self):
        return "MathBlock(%s, %s, %s, %s)" % (repr(self._id), repr(self._name), repr(self._complexity), repr(self._notation))


def _from_json(body):
    bloc = MathBlock(None, None, None, None)
    bloc._from_json(body)
    return bloc




# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from search import Search
from variable_details import VariableDetails
from utils import *
from utils.jsonrest import _JsonREST

from session import Http404Exception

import base64
import json
import urllib

import warnings

class DataSource(_JsonREST):
    """Acts as an interface to a data source on the server.

    DataSources can be created by calling :py:meth:`~eureqa.eureqa.Eureqa.create_data_source`
    or an existing one can be retreived with :py:meth:`~eureqa.eureqa.Eureqa.get_data_source`

    :var str `~eureqa.data_source.DataSource.name`: The data source name.
    :var str `~eureqa.data_source.DataSource.series_id_column_name`: The name of the column that splits data into series based on its values.
    :var str `~eureqa.data_source.DataSource.series_order_column_name`: The name of the column that defines the order of the data. This indicates the data is timeseries data and will sort the rows based on this column.
    :var int `~eureqa.data_source.DataSource.number_columns`: The number of columns (variables) in the data source.
    :var int `~eureqa.data_source.DataSource.number_rows`: The number of rows in the data source.
    :var int `~eureqa.data_source.DataSource.number_series`: The number of series (chunks of rows) in the data source.
    """

    EXISTING_ROW_ORDER = "<row>"

    def __init__(self, eureqa, datasource_name=None, hidden=False, datasource_id=None):
        """For internal use only
        PARAM_NOT_EXTERNALLY_DOCUMENTED
        """

        super(DataSource, self).__init__(eureqa)

        if datasource_id is not None:
            self._datasource_id = datasource_id
        if datasource_name is not None:
            self._datasource_name = datasource_name

        self._hidden = hidden
        self._series_count = 0
        self._series_descriptions = []
        self._prepared_data_col_count = 0
        self._prepared_data_row_count = 0
        self._file_name = ''
        self._file_size = 0
        self._file_uploaded_user = ''
        self._file_uploaded_date = ''
        self._series_order_column = None
        self._series_id_column = None

    def _directory_endpoint(self):
        return "/fxp/datasources"
    def _object_endpoint(self):
        return "/fxp/datasources/%s" % (self._datasource_id)
    def _fields(self):
        return ["datasource_id", "datasource_name", "prepared_data_row_count", "prepared_data_col_count", "series_count", "series_descriptions",  "file_name", "file_size", "file_uploaded_user", "file_uploaded_date", "hidden", "series_order_column", "series_order_variable", "series_id_column"]

    @property
    def name(self):
        return self._datasource_name
    @name.setter
    def name(self, val):
        self._datasource_name = val
        self._update()

    @property
    def number_columns(self):
        return self._prepared_data_col_count

    @property
    def number_rows(self):
        return self._prepared_data_row_count

    @property
    def number_series(self):
        return self._series_count

    @property
    def _data_source_id(self):
        return self._datasource_id

    @property
    def _data_file_name(self):
        return self._file_name

    @property
    def _data_file_size(self):
        return self._file_size

    @property
    def _data_file_uploaded_user(self):
        return self._file_uploaded_user

    @property
    def _data_file_uploaded_date(self):
        return self._file_uploaded_date

    @property
    def series_order_column_name(self):
        """ The name of the column that defines the order of the data. This indicates the data is timeseries data and will sort the rows based on this column. Use DataSource.EXISTING_ROW_ORDER to use the current row number as the series order.

        """
        return self._series_order_column
    @series_order_column_name.setter
    def series_order_column_name(self, val):
        self._update_metadata(series_id_column_name=self.series_id_column_name, series_order_column_name=val)

    @property
    def series_id_column_name(self):
        """The name of the column that splits data into series based on its values.

        """
        return self._series_id_column

    @property
    def series_order_variable_name(self):
        """The name of the variable that defines the order of the data.  Same as `series_order_column_name` except properly escaped for use as a variable name.

        """
        return self._series_order_variable

    @series_id_column_name.setter
    def series_id_column_name(self, val):
        self._update_metadata(series_id_column_name=val, series_order_column_name=self.series_order_column_name)

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    def delete(self):
        """Deletes the data source from the server.

        :raise Exception: If the data source is already deleted.
        """
        self._delete()

    def _create(self, file_or_path):
        assert self._datasource_name is not None, "Cannot create DataSource without a name"
        assert not hasattr(self, "_datasource_id"), "Cannot specify DataSources' ID before creation"

        if isinstance(file_or_path, basestring):
            f = open(file_or_path, 'rb')
            file_path = file_or_path
        else:
            f = file_or_path
            file_path = getattr(file_or_path, 'name', str(file_or_path))

        try:
            raw_data = {}
            if getattr(self, "_series_id_column", None):
                raw_data["series_id_column"] = self._series_id_column
            if getattr(self, "_series_order_column", None):
                raw_data["series_order_column"] = self._series_order_column
            raw_data['datasource_name'] = self.name
            if self._hidden:
                raw_data['hidden'] = "true"
            else:
                raw_data['hidden'] = "false"
            resp = self._eureqa._session.execute('/fxp/datasources/create_and_upload', 'POST', files={'file': (file_path,f)}, raw_data=raw_data)
            self._from_json(resp.get('datasource'))
        finally:
            if isinstance(file_or_path, basestring):
                f.close()
        self._get_self()

    def _update_metadata(self, series_id_column_name, series_order_column_name):
        self._series_id_column = series_id_column_name
        self._series_order_column = series_order_column_name
        self._update()

    def download_data_file(self, file_path):
        """Downloads the originally uploaded data from the server.

        :param str file_path: the filepath at which to save the data

        """
        result = self._eureqa._session.execute('/fxp/datasources/%s/download' % self._data_source_id, 'GET', raw_returnfile=file_path)

    def get_variables(self, _derivation_type = 'all'):
        """Retrieves from the server a list of variables in a datasource.

        :return:
            A list of the same variables as visible in Eureqa UI.
            Including all derived variables.
        :rtype: list of str
        """
        allowed_derivation_type = ['all', 'original', 'custom', 'seasonal']
        if _derivation_type not in allowed_derivation_type:
            raise ValueError("The argument 'derivation_type' must be one of " + str(allowed_derivation_type))
        endpoint = '/fxp/datasources/%s/variables?sort=[{"key":"index"}]' % self._datasource_id
        self._eureqa._session.report_progress('Getting variable details for datasource: \'%s\'.' % self.name)
        body = self._eureqa._session.execute(endpoint, 'GET')
        if _derivation_type == 'all':
            return [x['variable_name'] for x in body]
        else:
            return [x['variable_name'] for x in body if x['type'] == _derivation_type]

    def get_searches(self):
        """Retrieves from the server a list of searches associated with the data source.

        :return: The list of all searches associated with the data source.
        :rtype: list of :class:`~eureqa.search.Search`
        """

        endpoint = '/fxp/datasources/%s/searches' % self._datasource_id
        self._eureqa._session.report_progress('Getting searches for datasource: \'%s\'.' % self.name)
        body = self._eureqa._session.execute(endpoint, 'GET')
        return [Search(x, self._eureqa, self) for x in body]

    def create_search(self, search_settings, _hidden = False):
        """Creates a new search with settings from a :any:`SearchSettings` object.

        :param SearchSettings search_settings: the settings for creating a new search.
        :return: A :class:`~eureqa.search.Search` object which represents a newly create search on the server.
        :rtype: ~eureqa.search.Search
        """

        endpoint = "/fxp/datasources/%s/searches" % self._datasource_id
        body = search_settings._to_json()
        body['hidden'] = _hidden
        self._eureqa._session.report_progress('Creating search for datasource: \'%s\'.' % self.name)
        result = self._eureqa._session.execute(endpoint, 'POST', body)
        search_id = result['search_id']
        return self._eureqa._get_search_by_search_id(self._datasource_id, search_id)

    def evaluate_expression(self, expressions, _data_split='all'):
        warnings.warn("This function has been deprecated.  Please use `Eureqa.evaluate_expression()` instead.", DeprecationWarning)
        return self._eureqa.evaluate_expression(self, expressions, _data_split=_data_split)

    def get_series_id_values(self):
        """Get all available series id values for this dataset. Note the series id for a single series dataset is reported to be ''

        :return: a list of all available series_id values
        :rtype: list
        """
        return [entry["series_id_val"] for entry in self._series_descriptions]

    def _compute_series_indicator(self):
        """For internal use only.

        Compute an indicator column which indicates the series which each row belongs to, as an integer from 0 to (num_series-1).

        This code was written with the assumption that it would be applied to the entire datasource, as opposed to the training or validation split.
        """
        indicator = []
        for i,series_desc in enumerate(self._series_descriptions):
            indicator += [i]*series_desc['series_size']
        return indicator

    def create_variable(self, expression, variable_name):
        """Adds a new variable to the :py:class:`~data_source.DataSource` with values from evaluating the given expression.

        :param str expression: the expression to evaluate to fill in the values
        :param str variable_name: what to name the new variable
        """
        endpoint = '/fxp/datasources/%s/variables' % self._datasource_id
        body = {'datasource_id': self._datasource_id,
                'expression': expression,
                'variable_name': variable_name}
        result = self._eureqa._session.execute(endpoint, 'POST', body)
        self.__dict__ = self._eureqa.get_data_source_by_id(self._datasource_id).__dict__

    def get_variable_details(self, variable_name):
        """Retrieves the details for the requested variable from the :py:class:`~data_source.DataSource`.

        :param str variable_name: the name of the variable to get the details for
        :return:
            The object representing the variable details
        :rtype: VariableDetails
        """
        endpoint = '/fxp/datasources/%s/variables/%s' % (self._datasource_id, urllib.quote_plus(base64.b64encode('%s-_-%s' % (self._datasource_id, variable_name))))
        self._eureqa._session.report_progress('Getting variable details for datasource: \'%s\'.' % self.name)
        body = self._eureqa._session.execute(endpoint, 'GET')
        return VariableDetails(body, self)

    def get_variable(self, variable_name):
        warnings.warn("'get_variable()' function deprecated; please call as 'get_variable_details()' instead", DeprecationWarning)
        return self.get_variable_details(variable_name)

    def _create_seasonality_variable(self, seasonal_target_variable, seasonal_period):
        """Adds a seasonality variable to the data_source with specified target variable and seasonal period

        :param str seasonality_target_variable: the variable name to calculate the seasonality for
        :param str seasonal_period: period of the seasonal effect, supported options are 'yearly', 'weekly' or 'daily'
        :return: name of the newly created variable
        :rtype: str
        """
        endpoint = '/fxp/datasources/%s/variables' % self._datasource_id
        body = {'datasource_id': self._datasource_id,
                'type': 'seasonal',
                'seasonal_target_variable': seasonal_target_variable,
                'seasonal_period': seasonal_period}
        result = self._eureqa._session.execute(endpoint, 'POST', body)
        self.__dict__ = self._eureqa.get_data_source_by_id(self._datasource_id).__dict__
        return result['variable_name']

    def _create_variable_from_template(self, template):
        """Adds a custom variable to the data_source with the expression provided in the template of variable details

        :param VariableDetails template: the template to use for creating the new variable
        """
        if template.derivation_type == 'original':
            raise ValueError('Original variables cannot be used as the template')

        self.create_variable(template.expression, template.name)

    def _apply_solutions(self, solutions, expressions=[]):
        solution_requests = []
        for solution in solutions:
            solution_request = {
                    "datasource_id": solution._datasource_id,
                    "search_id": solution._search_id,
                    "solution_id": solution._id}
            solution_requests.append(solution_request);
        body = {
            "solutions": solution_requests,
            "expressions": expressions}
        endpoint = '/fxp/datasources/%s/apply_solutions' % self._datasource_id
        result = self._eureqa._session.execute(endpoint, 'POST', body)
        return self._eureqa._get_search_by_search_id(result["datasource_id"], result["search_id"])

    def _duplicate(self, new_name, hidden=False):
        endpoint = '/fxp/datasources/%s/duplicate' % self._datasource_id
        body = {'datasource_name': new_name,
                'hidden': hidden}
        result = self._eureqa._session.execute(endpoint, 'POST', body)
        duplicate = DataSource(self._eureqa)
        duplicate._from_json(result)
        return duplicate




# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


class ComplexityWeights:
    """Represents complexity penalties applied to each model to reward for increased interpretability.

    :param int continuous_variable: additional complexity added for each continuous variable in a term after the first continuous variable.
    :param int boolean_variable: additional complexity added for each binary variable in a model.
    :param int nest_depth: additional complexity added for each term with nested functions, like log(sin(x)).
    """

    def __init__(self, continuous_variable=2, boolean_variable=0, nest_depth=0):
        """For internal use only.
        :param int continuous_variable: additional complexity added for each continuous variable in a term after the first continuous variable.
        :param int boolean_variable: additional complexity added for each binary variable in a model.
        :param int nest_depth: additional complexity added for each term with nested functions, like log(sin(x)).
        """
        self.nest_depth = nest_depth
        self.boolean_variable = boolean_variable
        self.continuous_variable = continuous_variable

    def _from_json(self, body):
        self.nest_depth = body['nest_depth_weight']
        self.boolean_variable = body['num_boolean_variable_weight']
        self.continuous_variable = body['num_continuous_variable_weight']
        self._body = body

    @classmethod
    def from_json(cls, body):
        complexityWeights = ComplexityWeights()
        complexityWeights._from_json(body)
        return complexityWeights

    def _to_json(self):
        body = {'nest_depth_weight': self.nest_depth,
                'num_boolean_variable_weight': self.boolean_variable,
                'num_continuous_variable_weight': self.continuous_variable}
        return body

    def __eq__(self, other):
        if not isinstance(other, ComplexityWeights):
            return False
        return (self.nest_depth == other.nest_depth and self.boolean_variable == other.boolean_variable
                and self.continuous_variable == other.continuous_variable)

    def __ne__(self, other):
        return not (self == other)


def _from_json(body):
    weights = ComplexityWeights()
    weights._from_json(body)
    return weights



from eureqa.analysis.components.base import _Component
from eureqa.utils.jsonrest import _JsonREST

class MagnitudeBar(_Component):
    """
    Magnitude Bar: implements a visual representation of percentages

    For Example::

        pink_bar = MagnitudeBar(value=-0.22, color='#ff00ff')
        analysis.create_html_card("Here is the magnitude bar: {0}".format(analysis.html_ref(pink_bar)))

    :param float value: value expressed by this MagnitudeBar
    :param str color: (optional) html hex color code to use for the magnitude bar instead of default
    """

    _component_type_str = 'MAGNITUDE_BAR'

    def __init__(self, value=None, color=None, _analysis=None, _component_id=None, _component_type=None):
        if value is not None:
            self._value = value

        if color is not None:
            self._color = color

        super(MagnitudeBar, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)

    @property
    def value(self):
        """The value expressed by this MagnitudeBar.
        Must be a fractional number between 0 and 1 (positive bar) or 0 and -1 (negative bar).

        :rtype: float
        """
        return getattr(self, "_value", None)

    @value.setter
    def value(self, val):
        self._value = val
        self._update()

    @property
    def color(self):
        """The Color of this bar, expressed as an html hex color code.

        For example: '#ff5733' - red

        :rtype: str
        """
        return getattr(self, "_color", None)

    @color.setter
    def color(self, val):
        self._color = val
        self._update()

    def _fields(self):
        return super(MagnitudeBar, self)._fields() + [ 'value', 'color' ]



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from variable_options import VariableOptions

class VariableOptionsDict(dict):
    """
    Create a dictionary of :py:class:`~eureqa.variable_options.VariableOptions` objects
    in which the key is the :py:class:`~eureqa.variable_options.VariableOptions` name.
    
    This class is used primarily to hold variable parameter overrides for 
    :py:class:`~eureqa.eureqa.Eureqa` and :py:class:`~eureqa.search.Search` .
    """
    def add(self, variable_options):
        """
        Add an existing :py:class:`~eureqa.variable_options.VariableOptions` to this dictionary.

        :param VariableOptions variable_options:  :py:class:`~eureqa.variable_options.VariableOptions` object to add
        """
        assert isinstance(variable_options, VariableOptions), "VariableOptionsDict can only hold VariableOptions objects.  Tried to insert an object of type '%s'." % variable_options.__class__.__name__
        self[variable_options.name] = variable_options
    
    def __missing__(self, key):
        v = VariableOptions(name = key)
        self[key] = v
        return v 

    def __setitem__(self, key, value):
        ## Override __setitem__ to do some validation
        assert isinstance(value, VariableOptions), "VariableOptionsDict can only hold VariableOptions objects.  Tried to insert an object of type '%s'." % value.__class__.__name__
        assert key == value.name, "Can't assert VariableOption named '%s' into VariableOptionsDict field '%s':  Names must match" % (value.name, key)
        assert key not in self, "VariableOptionsDict already has a VariableOptions for variable %s" % (key)
        super(VariableOptionsDict, self).__setitem__(key, value)

    def _to_json(self):
        return [x._to_json() for x in self.itervalues()]

    def _from_json(self, vars):
        for var in vars:
            var = VariableOptions._from_json(var)
            self[var.name] = var

    @classmethod
    def from_json(cls, vars):
        vod = cls()
        vod._from_json(vars)
        return vod
    



from eureqa.analysis.components.base import _Component

class TitledLayout(_Component):
    """A layout with space for a title and description and content within an Analysis

    For example::

        h=HtmlBlock(html="<h1>Hybrid Performance</h1>")
        layout=TitledLayout(title="the title", description="You can add content specific description here", content=h)
        analysis.create_card(layout)

    :param str title: The title of the layout
    :param str description: The description of the layout (can contain HTML)
    :param _Component content: The component to use in the main content of the layout

    """

    _component_type_str = 'TITLED_LAYOUT'

    def __init__(self, title=None, description=None, content=None, _analysis=None, _component_id=None, _component_type=None):
        super(TitledLayout, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)

        if title is not None:
            self._title = title

        if description is not None:
            self._description = description

        if content is not None:
            self.create_card(content)


    @property
    def title(self):
        """The title of this card.

        :return: title of this card
        :rtype: str"""
        return self._title

    @title.setter
    def title(self, val):
        self._title = val
        self._update()


    @property
    def description(self):
        """ The description of this card.

        :return: description of this card
        :rtype: str"""
        return self._description

    @description.setter
    def description(self, val):
        self._description = val
        self._update()


    @property
    def content(self):
        """The Component that this TitledLayout is adding a title to.
        This field can't be assigned to; use 'create_card()' to
        assign a Component to this TitledLayout.

        :rtype: ~eureqa.analysis.components._Component"""
        if not hasattr(self, "_content_component"):
            self._content_component = _Component._get(_analysis=self._analysis, _component_id=self._content_component_id)
        return self._content_component

    def create_card(self, component):
        """
        Assign a Component to this TitledLayout

        :param _Component component: Component to use as the content of this layout
        """
        if hasattr(self, "_analysis"):
            component._associate(self._analysis)

        # Set all local fields that we're able to set at this time
        self._content_component = component
        if hasattr(self, "_content_component_id"):
            del self._content_component_id
        if hasattr(component, "_component_id"):
            self._content_component_id = component._component_id

        # Save any updates to ourself
        self._update()

    def delete(self):
        if hasattr(self, "_content_component"):
            self._content_component.delete()
        super(TitledLayout, self).delete()

    def clone(self):
        clone = super(TitledLayout, self).clone()
        clone.create_card(self.content.clone())
        return clone

    def _walk_children(self):
        if hasattr(self, "_content_component"):
            yield self._content_component

    def _register(self, analysis):
        super(TitledLayout, self)._register(analysis)
        if hasattr(self, "_content_component"):
            self.create_card(self._content_component)

    def _associate_with_table(self, analysis):
        child_json = []
        if hasattr(self, "_content_component") and not hasattr(self._content_component, "_component_id"):
            children = self._content_component._associate_with_table(analysis)
            self.create_card(self._content_component)
            child_json += children
        return super(TitledLayout, self)._associate_with_table(analysis) + child_json

    def _fields(self):
        return super(TitledLayout, self)._fields() + [ 'title', 'description', 'content_component_id' ]



 # Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import glob, os
import time
from functools import wraps

def remove_files_in_directory(dir_path):
    """Removes all files in the specified directory.

    :param str dir_path: Path to files to be removed.
    """
    filelist = glob.glob(os.path.join(dir_path, '*'))
    for f in filelist:
        try:
            os.remove(f)
        except Exception as e:
            print 'Warning: remove_files_in_directory received exception "%s"' % e

class Throttle(object):
    """Decorator that ensures the wrapped function is not called twice within one seconds.
    sleeps for a second if it is called within 1 second of the last call
    must wrap a function that takes at least one argument"""
    def __init__(self):
        # keys are __str__ representation of first argument, values
        # are last call time b/c there is once Throttle object
        # instance for each decorated function (not for each object
        # instance with a decorated function)
        self._last_call_time = {}

    def __call__(self, wrapped):
        @wraps(wrapped)
        def wrapper(*args, **kwargs):
            this_call_time = time.time()
            # pick string representation of the first arg (aka self)
            key = id(args[0])
            # delay the call by a second if it has been less than
            if key in self._last_call_time and this_call_time - self._last_call_time[key] < 1:
                time.sleep(1)

            # Important to set last call to the time this method was called. Not after the sleep.
            # Otherwise it will get into the infinite loop.
            self._last_call_time[key] = this_call_time

            # do the actual work
            return wrapped(*args, **kwargs)

        return wrapper



#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from parameter import Parameter
from variable_parameter import VariableParameter

import json

class DataSourceParameter(Parameter):
    """Data source parameter description for analysis template

    :param str id: The id of the parameter that will be passed together with its value to an analysis script.
    :param str label: The parameter label that will be shown in UI.
    :param list[eureqa.analysis_templates.VariableParameter] variables: The parameters for variables that belong to the datasource described
        by this parameter.
    """

    def __init__(self, id, label, variables):
        """Initializes a new instance of the ~DataSourceParameter class
        :param str id: The id of the parameter that will be passed together with its value to an analysis script.
        :param str label: The parameter label that will be shown in UI.
        :var list[eureqa.analysis_templates.VariableParameter] variables: The parameters for variables that belong to the datasource described
        by this parameter.
        """
        Parameter.__init__(self, id, label, "datasource")
        self.variables = variables

    def _to_json(self):
        body = Parameter._to_json(self)
        if self.variables is not None and len(self.variables) > 0:
            body['parameters'] = [x._to_json() for x in self.variables]
        return body

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    @staticmethod
    def _from_json(body):
        ret = DataSourceParameter(None, None, None)
        Parameter.from_json(ret, body)
        if ret._type != "datasource": raise Exception("Invalid type '%s' specified for datasource parameter" % ret._type)
        ret.variables = []
        if 'parameters' in body:
            ret.variables = [VariableParameter._from_json(x) for x in body['parameters'] if 'type' in x and x['type'] == "variable"]
        return ret



import os
import sys
import json
import argparse
import textwrap

from eureqa.analysis_templates.runner import analysis_template_runner
from eureqa.utils._analysis_template_cli import *

def main(argv):
    """ Handle executing analysis template code without actually uploading it to the server.
    Mostly a wrapper around :class:`~eureqa.analysis_templates.runner.analysis_template_runner`.
    The latter is called directly by the server; this one is a little more user-friendly
    and respects eureqa_config.json.

    :param list argv: Arguments
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=DESCRIPTION, epilog=textwrap.dedent("""\
        Usage Examples:

        To run the example module included with this project, do:
        python -m eureqa run_analysis_template example_module
        You will be prompted for username and password if necessary.

        To run within a particular Eureqa Organization:
        python -m eureqa run_analysis_template -u <username> -p <password> -o <organization> example_module
        """))

    ## Pick the command-line arguments that this app needs
    add_module_args(parser)
    add_server_args(parser)
    add_local_template_args(parser)
    add_script_args(parser)

    ## parse_args() wants the arguments to the command.
    ## argv[0] *is* the command.
    args = parser.parse_args(argv[1:])

    params = merge_defaults_from_json(args.module_dir, args)

    clean_and_populate_params(params, not args.quiet)
    module_dir  = params['module_dir']
    module_name = os.path.basename(module_dir)

    if not args.quiet:
        params_no_password = dict(params)
        params_no_password['password'] = '********'

        # Flatten structure for pretty-printing
        if 'parameter_values' in params_no_password and isinstance(params_no_password['parameter_values'], dict):
            params_no_password['parameter_values'] = params_no_password['parameter_values']['parameters']

        print "Installing module '%s':\n%s" % (module_name, json.dumps(params_no_password,
                                                                      indent=4,
                                                                      sort_keys=True))

    runner_args = ['-m', module_name,
                   '-u', params['username'],
                   '-e', params['url'],
                   '-o', params['organization']]

    if 'password' in params:
        runner_args.append('-p')
        runner_args.append(params['password'])
    if 'api_key' in params:
        runner_args.append('-k')
        runner_args.append(params['api_key'])
    if 'save_credentials' in params:
        runner_args.append('--save-credentials')
    if 'quiet' in params:
        runner_args.append('--quiet')

    ## Make sure the module is in path so it can be imported
    mod_path = os.path.dirname(os.path.abspath(params['module_dir']))
    sys.path.insert(0, mod_path)

    runner = analysis_template_runner()
    runner.run_args(runner_args, params.get('parameter_values', {}))

    if args.save:
        write_defaults_to_json(params, module_dir)

    sys.path.remove(mod_path)



#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from parameter import Parameter

import json

class NumericParameter(Parameter):
    """Numeric parameter description for analysis template

    :param str id: The id of the parameter that will be passed together with its value to an analysis script.
    :param str label: The parameter label that will be shown in UI.
    :param float min: The minimum value to allow.
    :param float max: The maximum value to allow.
    :param bool require_int: Indicates that the number should be an integer.

    :var str id: The id of the parameter that will be passed together with its value to an analysis script.
    :var str NumericParameter.label: The parameter label that will be shown in UI.
    :var float min: The minimum value to allow.
    :var float max: The maximum value to allow.
    :var bool reqire_int: Indicates that the number should be an integer.
    """

    def __init__(self, id, label, min=None, max=None, require_int=False):
        """Initializes a new instance of the ~TextParameter class
        """
        Parameter.__init__(self, id, label, "numeric")
        self.min = min
        self.max = max
        self.require_int = require_int

    def _to_json(self):
        body = Parameter._to_json(self)
        if self.min is not None:
            body["min"] = self.min
        if self.max is not None:
            body["max"] = self.max
        if self.require_int is not None:
            body["require_int"] = self.require_int
        return body

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    @staticmethod
    def _from_json(body):
        param = NumericParameter(None, None)
        param.from_json(body)
        if param._type != "numeric": raise Exception("Invalid type '%s' specified for numeric parameter" % param._type)
        if 'min' in body:
            param.min = body["min"]
        if 'max' in body:
            param.max = body["max"]
        if 'require_int' in body:
            param.require_int = body["require_int"]
        return param



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from eureqa.analysis.components.base import _Component
from eureqa.data_source import DataSource

class ModelFitByRowPlot(_Component):
    """A model fit by row plot card. See also :any:`Analysis.create_model_fit_by_row_plot_card`

    For example::

        p = ModelFitByRowPlot(s.get_best_solution())
        analysis.create_card(p)


    :param Solution solution: The solution that will be displayed on the card.
    :param bool use_all_data: If true, uses all data int he datasource. Otherwise use training set
    """

    _component_type_str = 'PROJECTION_CARD'

    def __init__(self, solution=None, use_all_data=None, _analysis=None, _component_id=None,
                 _component_type=None, **kwargs): #**kwargs - for backward compatibility
        if isinstance(solution, DataSource):
            # This is backward compatibility hack to support the
            # (self, datasource, search, solution, use_all_data) calling signature.
            solution = _analysis
            use_all_data = _component_id

            # Only support backwared compatible signature for public params
            # because the internal calls were ported to new signature.
            _analysis = None
            _component_id = None
            _component_type = None

        if solution is not None:
            self.solution = solution

        if use_all_data is not None:
            self.use_all_data = use_all_data

        super(ModelFitByRowPlot, self).__init__(_analysis=_analysis, _component_id=_component_id, _component_type=_component_type)

    @property
    def solution(self):
        """The Solution that is being explained

        :rtype: eureqa.Solution
        """
        if hasattr(self, "_eureqa"):
            return self._eureqa._get_solution_by_id(self._datasource_id, self._search_id, self._solution_id)

    @solution.setter
    def solution(self, val):
        self._solution_id = val._id
        self._search_id = val._search_id
        self._datasource_id = val._datasource_id
        self._eureqa = val._eureqa
        self._update()

    @property
    def use_all_data(self):
        """Use all data or just validation data?

        :rtype: bool
        """
        return getattr(self, "_use_all_data", None)

    @use_all_data.setter
    def use_all_data(self, val):
        self._use_all_data = val
        self._update()

    def _fields(self):
        return super(ModelFitByRowPlot, self)._fields() + [ 'datasource_id', 'search_id', 'solution_id', 'use_all_data' ]



from eureqa.utils.jsonrest import _JsonREST
from eureqa.analysis.components.base import _Component

class Card(_JsonREST):
    """The initially visible content of an Analysis is defined as a list
    of :class:`Card` s. Cards are ordered and can be collapsed by default, and
    each has a default Component which is visble.

    You should not create a :class:`Card` manually, but rather use
    :meth:`~eureqa.analysis.Analysis.create_card`

    :var bool collapse:  If the Card is currently collapsed or not

    """

    @property
    def collapse(self):
        """ Whether this item should be rendered as "collapsed" (with its content hidden)

        :type: bool
        """
        return self._collapse

    @collapse.setter
    def collapse(self, val):
        self._collapse = val
        self._update()

    def __init__(self,
                 _analysis=None, _item_id=None, _component=None,
                 _order_index=None, _collapse=None):
        """ For internal use only """

        if _analysis is not None:
            _JsonREST.__init__(self, _analysis._eureqa)
            self._analysis = _analysis

        if _component is not None:
            self.component = _component

        if _item_id is not None:
            self._item_id = _item_id
        if _order_index is not None:
            self._order_index = _order_index
        if _collapse is not None:
            self._collapse = _collapse

        if _analysis is not None and _component is not None:
            self._associate(_analysis)

    def _associate(self, analysis):
        _JsonREST.__init__(self, analysis._eureqa)
        self._analysis = analysis
        self._analysis_id = analysis._id
        if getattr(self, "_default_component", None):
            self._default_component._associate(analysis)
            self._default_component_id = self._default_component._component_id
        # Create a matching object on the server
        self._create()

    def _fetch_component(self):
        if not getattr(self, "_default_component", None) and getattr(self, "_analysis", None):
            self.component = _Component._get(
                _component_id=self._default_component_id,
                _analysis=self._analysis)

    @property
    def component(self):
        return self._default_component

    @component.setter
    def component(self, val):
        assert not getattr(val, "_analysis", None) or not getattr(self, "_analysis", None) or val._analysis._id == self._analysis._id, \
            "Can't attach component to analysis '%s'; already attached to analysis '%s'" % (self._analysis._id, val._analysis._id)

        if getattr(self, "_analysis", None):
            val._associate(self._analysis)

        self._default_component = val
        if getattr(val, "_component_id", None):
            self._default_component_id = val._component_id
        self._update()

    # Wrap our contained Component's fields for backwards compatibility.
    #
    # Note that __getattr__ is called only if normal object access fails, but
    # __setattr__ and __delattr__ replace normal object access completely so
    # must invoke it explicitly as a base case.
    # (Python works this way as a common-case performance optimization.)
    def __getattr__(self, name):
        if "_default_component" not in self.__dict__:
            raise AttributeError("Field %s not found" % repr(name))
        component = super(Card, self).__getattribute__("_default_component")
        try:
            return getattr(component, name)
        except AttributeError, e:
            if component._component_type == "TITLED_LAYOUT":
                return getattr(component.content, name)
            raise e

    def __setattr__(self, name, value):
        if "_default_component" in self.__dict__:
            component = self.__dict__["_default_component"]
            if hasattr(component, name):
                return setattr(component, name, value)
            if component._component_type == "TITLED_LAYOUT" and \
                hasattr(component.content, name):
                return setattr(component.content, name, value)
        super(Card, self).__setattr__(name, value)

    def __delattr__(self, name):
        if "_default_component" in self.__dict__:
            component = self.__dict__["_default_component"]
            if name in component.__dict__:
                return delattr(component, name)
            if component._component_type == "TITLED_LAYOUT" and \
                name in component.content.__dict__:
                return delattr(component.content, name)
        super(Card, self).__delattr__(name)

    # List of fields on this object.
    # Typically used by code-autocomplete tools.
    def __dir__(self):
        fields = []
        fields += self.__dict__.keys()
        if hasattr(self, "_default_component"):
            fields += self._default_component.__dict__.keys()
            if hasattr(self._default_component, "_content_component"):
                fields += self._default_component.content.__dict__.keys()

        return sorted(set(self.__dict__.keys()))

    def _object_endpoint(self):
        return '/analysis/%s/items/%s' % (self._analysis._id, self._item_id)
    def _directory_endpoint(self):
        return '/analysis/%s/items' % (self._analysis._id)
    def _fields(self):
        return [ "analysis_id", "item_id", "default_component_id", "order_index", "collapse" ]

    def __repr__(self):
        try:
            component = self.component
        except AttributeError:
            component = getattr(self, "_default_component_id", "??")
        return "Card(component=%s, _order_index=%s, _collapse=%s, _item_id=%s, _analysis_id=%s)" % \
               (repr(getattr(self, "component", None)), repr(getattr(self, "_order_index", None)),
                repr(getattr(self, "collapse", None)), repr(getattr(self, "_item_id", None)),
                repr(getattr(self, "_analysis_id", None)))

    def _to_json(self):
        resp = super(Card, self)._to_json()

        if hasattr(self, "_analysis"):
            resp["analysis_id"] = self._analysis._id

        return resp

    def _from_json(self, body):
        if hasattr(self, "_analysis") and body.get("analysis_id"):
            assert body.get("analysis_id") == self._analysis._id, \
                "_from_json() can't de-serialize an Item that belongs to a different Analysis"


        super(Card, self)._from_json(body)

    @classmethod
    def _construct_from_json(cls, body, *args, **kwargs):
        ret = super(Card, cls)._construct_from_json(body, *args, **kwargs)
        ret._fetch_component()
        return ret

    def clone(self):
        # Make a copy of ourselves
        body = self._to_json()

        # Disassociate the copy from the current Analysis
        if "analysis_id" in body.keys(): del body["analysis_id"]
        if "item_id" in body.keys(): del body["item_id"]

        # Instantiate a new Item with this state
        new_item = Card._construct_from_json(body)

        # Handle Components
        if hasattr(self, "_default_component"):
            new_item.component = self.component.clone()

        return new_item

    def delete(self):
        if hasattr(self, "_default_component"):
            self._default_component.delete()
        self._delete()

    def move_above(self, other_card):
        """Moves this card above another card.

        :param Card other_card: The other card above which to move this card.
        """

        other_card_order = other_card._order_index if hasattr(other_card, '_order_index') else other_card
        if other_card_order >= self._order_index:
            return
        self._order_index = other_card_order
        self._update()

    def move_below(self, other_card):
        """Moves this card below another card.

        :param Card other_card: The other card object below which to move this card.
        """

        other_card_order = other_card._order_index if hasattr(other_card, '_order_index') else other_card
        if other_card_order <= self._order_index:
            return
        self._order_index = other_card_order
        self._update()



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from search_settings import SearchSettings
from variable_options_dict import VariableOptionsDict
from variable_options import VariableOptions

class SearchTemplates:
    """Provides a set of search settings for well known search scenarios.
    Created automatically when you create a :py:class:`~eureqa.eureqa.Eureqa`. Can be accessed through :py:class:`~eureqa.eureqa.Eureqa.search_templates`.

    Each function of this class represents a type of search that there is a template for.
    The function will return a :py:class:`~eureqa.search_settings.SearchSettings`
    containing the suggested settings for that type of search.

    """

    Numeric                  = "generic"
    Timeseries               = "timeseries"
    Classification           = "classification"
    ClassificationTimeseries = "timeseries_classification"

    def __init__(self, eureqa):
        """For internal use only
        PARAM_NOT_EXTERNALLY_DOCUMENTED
        """

        self._eureqa = eureqa

    def numeric(self, name, target_variable, input_variables, variable_options=None):
        """The numeric search settings template.

        :param str name: The search name.
        :param str target_variable: The target variable.
        :param list input_variables: The list (str) of input variables.
        :param list variable_options: An optional list of :py:class:`~eureqa.variable_options.VariableOptions`
            to include in the search
        :rtype: SearchSettings
        """

        assert not isinstance(input_variables, str), "Input variables must be a list."
        try: input_variables = list(input_variables) # make sure the input variables are a list
        except TypeError: raise Exception('Input variables must be a list.')

        endpoint = '/fxp/search_templates/%s/create_settings' % self.Numeric
        args = {
            'search_name': name,
            'input_variables': input_variables,
            'target_variable': target_variable,
            'search_template_id': self.Numeric
        }
        self._add_variable_options_to_args(args, variable_options)
        self._eureqa._session.report_progress('Creating numeric settings for search: \'%s\'.' % name)
        body = self._eureqa._session.execute(endpoint, 'POST', args)
        settings = SearchSettings.from_json(self._eureqa, body)
        return settings

    def classification(self, name, target_variable, input_variables, variable_options=None):
        """The classification search settings template.

        :param str name: The search name.
        :param str target_variable: The target variable.
        :param list input_variables: The list (str) of input variables.
        :param list variable_options: An optional list of :py:class:`~eureqa.variable_options.VariableOptions`
            to include in the search
        :rtype: SearchSettings
        """

        assert not isinstance(input_variables, str), "Input variables must be a list."
        try: input_variables = list(input_variables) # make sure the input variables are a list
        except TypeError: raise Exception('Input variables must be a list.')

        endpoint = '/fxp/search_templates/%s/create_settings' % self.Classification
        args = {
            'search_name': name,
            'input_variables': input_variables,
            'target_variable': target_variable,
            'search_template_id': self.Classification
        }
        self._add_variable_options_to_args(args, variable_options)
        self._eureqa._session.report_progress('Creating classification settings for search: \'%s\'.' % name)
        body = self._eureqa._session.execute(endpoint, 'POST', args)
        settings = SearchSettings.from_json(self._eureqa, body)
        return settings

    def time_series(self, name, target_variable, input_variables, min_delay=1, data_custom_history_fraction=0.1,
                    max_delays_per_variable=None, variable_options=None):
        """The time series search settings template.

        :param str name: The search name.
        :param str target_variable: The target variable.
        :param list input_variables: The list (str) of input variables.
        :param int min_delay: Optionally specify the minimum number of rows used in the range functions.
        :param float data_custom_history_fraction: Optionally specify the percentage of the data to be withheld
            from history blocks. Specifies the maximum possible delay for a history block.
        :param int max_delays_per_variable: Optionally overrides data_custom_history_fraction to directly set
            the maximum possible delay for a history block.
        :param list variable_options: An optional list of :py:class:`~eureqa.variable_options.VariableOptions`
            to include in the search
        :rtype: SearchSettings
        """

        assert not isinstance(input_variables, str), "Input variables must be a list."
        try: input_variables = list(input_variables) # make sure the input variables are a list
        except TypeError: raise Exception('Input variables must be a list.')

        endpoint = '/fxp/search_templates/%s/create_settings' % self.Timeseries
        args = {
            'search_name': name,
            'input_variables': input_variables,
            'target_variable': target_variable,
            'default_min_delay': min_delay,
            'max_delay': max_delays_per_variable,
            'search_template_id': self.Timeseries
        }
        self._add_variable_options_to_args(args, variable_options)
        self._eureqa._session.report_progress('Creating time series settings for search: \'%s\'.' % name)
        body = self._eureqa._session.execute(endpoint, 'POST', args)
        settings = SearchSettings.from_json(self._eureqa, body)
        return settings

    def time_series_classification(self, name, target_variable, input_variables, min_delay=1,
                                   data_custom_history_fraction=0.1, max_delays_per_variable=None, variable_options=None):
        """The time series classification search settings template.

        :param str name: The search name.
        :param str target_variable: The target variable.
        :param list input_variables: The list (str) of input variables.
        :param int min_delay: Optionally specify the number of rows the model should predict into the future (i.e. the default minimum number of rows each variable will be delayed in the model).
        :param float data_custom_history_fraction: Optionally specify the percentage of the data to be withheld
            from history blocks. Specifies the maximum possible delay for a history block.
        :param int max_delays_per_variable: Optionally overrides data_custom_history_fraction to directly set
            the maximum possible delay for a history block.
        :param list variable_options: An optional list of :py:class:`~eureqa.variable_options.VariableOptions`
            to include in the search
        :rtype: SearchSettings
        """

        assert not isinstance(input_variables, str), "Input variables must be a list."
        try: input_variables = list(input_variables) # make sure the input variables are a list
        except TypeError: raise Exception('Input variables must be a list.')

        endpoint = '/fxp/search_templates/%s/create_settings' % self.ClassificationTimeseries
        args = {
            'search_name': name,
            'input_variables': input_variables,
            'target_variable': target_variable,
            'default_min_delay': min_delay,
            'max_delay': max_delays_per_variable,
            'search_template_id': self.ClassificationTimeseries
        }
        self._add_variable_options_to_args(args, variable_options)
        self._eureqa._session.report_progress('Creating time series classification settings for search: \'%s\'.' % name)
        body = self._eureqa._session.execute(endpoint, 'POST', args)
        settings = SearchSettings.from_json(self._eureqa, body)
        return settings


    #Private Helper Methods
    def _add_variable_options_to_args(self, args, variable_options):
        if variable_options is not None:
            if isinstance(variable_options, VariableOptionsDict):
                args['variable_options'] = variable_options._to_json()
            else:
                try:
                    args['variable_options'] = []
                    for opt in variable_options:
                        assert isinstance(opt, VariableOptions)
                        args['variable_options'].append(opt._to_json())
                except TypeError, AssertionError:
                    raise Exception('variable_options must be a VariableOptionsDict or a list of VariableOptions')



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from execution import Execution
from parameters import Parameters
from eureqa.utils.modulefiles import module_path

import json
import tempfile
import zipfile
import os
import io

class AnalysisTemplate(object):
    """Represents an analysis template on the server.

    :var str `~eureqa.analysis_templates.analysis_template.name`: The name of the analysis template.
    :var str `~eureqa.analysis_templates.analysis_template.description`: The description of the analysis template.
    :var str `~eureqa.analysis_templates.analysis_template.icon_url`: The url of the icon for the analysis template.
    :var str `~eureqa.analysis_templates.analysis_template.icon_url_fill_custom`: If true, the icon_url will be filled with the analysis template's custom icon URL.
    :var list `~eureqa.analysis_templates.analysis_template.parameters`: The list of parameters for the analysis template.
    """

    def __init__(self, body, eureqa):
        """For internal use only
        PARAM_NOT_EXTERNALLY_DOCUMENTED
        """
        self._icon_url = body.get('icon_url')
        self._icon_url_fill_custom = body.get('icon_url_fill_custom')
        self._eureqa = eureqa
        self._body = body

    @property
    def _id(self):
        return self._body['id']
    # No setter -- don't change the ID; it's provided and opaque.
    # Also a private member.

    @property
    def name(self):
        """ Analysis Template's name """
        return self._body['name']
    @name.setter
    def name(self, val):
        self._body['name'] = val
        self._update_template()

    @property
    def description(self):
        """ Analysis Template's extended description """
        return self._body['description']
    @description.setter
    def description(self, val):
        self._body['description'] = val
        self._update_template()

    @property
    def parameters(self):
        """ Parameters object representing this template """
        return Parameters._from_json(self._body)
    @parameters.setter
    def parameters(self, val):
        if val is not None:
            self._body.update(val if isinstance(val, dict) else val._to_json())
        else:
            self._body['parameters'] = []
        self._update_template()

    def set_icon(self, icon_filepath):
        """
        Set the icon used for this analysis template

        :param str icon_filepath: path to a local file with a custom icon for the template.
        """
        if icon_filepath is None: return
        icon_filename = os.path.split(icon_filepath)[-1]
        f = open(icon_filepath, 'rb')
        res = self._eureqa._session.execute("/analysis_templates/%s/icon" % self._id,
                                            "POST", files={"file": (icon_filename, f)})
        self._icon_url_fill_custom = True
        self._update_template()

    def _update_template(self):
        endpoint = '/analysis_templates/%s' % (self._id)
        self._eureqa._session.report_progress('Updating analysis template: \'%s\'.' % (self._id))
        new_body = self._eureqa._session.execute(endpoint, 'POST', self._body)
        new_instance = self.__class__(new_body, self._eureqa)
        self.__dict__ = new_instance.__dict__

    def delete(self):
        """Delete the analysis template."""
        self._eureqa._session.report_progress('Deleting analysis_template: \'%s\'.' % self._id)
        self._eureqa._session.execute('/analysis_templates/%s' % (self._id), 'DELETE')

    def execute(self, values):
        """Start execution of an analysis template with given parameters.

        :param ~eureqa.analysis_templates.parameters_values.ParametersValues values: The parameter values to pass to the template.

        For example::

           ParametersValues(
               [TextParameterValue(eq, 'example_text_param','This is the text input value.'),
                DataSourceParameterValue(eq, '123', 'SimpleDataSource',
                   {VariableParameterValue(eq, '345', 'value 345')})],
           )
        """
        self._get_values_for_data_files(values.parameters)
        request = values._to_json()
        self._eureqa._session.report_progress('Creating execution for analysis_template: \'%s\'.' % self._id)
        result = self._eureqa._session.execute('/analysis_templates/%s/executions/' % (self._id), 'POST', request)
        return Execution(result, self._id, self._eureqa)

    def get_executions(self):
        """Get all executions of the analysis template."""
        self._eureqa._session.report_progress('Getting details for executions of analysis_template: \'%s\'.' % self._id)
        results = self._eureqa._session.execute('/analysis_templates/%s/executions' % (self._id), 'GET')
        return [Execution(x, self._id, self._eureqa) for x in results]

    def _get_execution(self, execution_id):
        """Get a specific execution of the analysis template.

        :param str execution_id: Execution identifier for this execution instance
        """
        self._eureqa._session.report_progress('Getting details of execution: \'%s\' of analysis_template: \'%s\'.' % (execution_id, self._id))
        result = self._eureqa._session.execute('/analysis_templates/%s/executions/%s' % (self._id, execution_id), 'GET')
        return Execution(result, self._id, self._eureqa)

    def set_module(self, main_module_name, module_fs_path, ignore_files=None, additional_modules_paths=[], icon_path=None):
        """Set the the python module and function that specifies the execution of this analysis template.

        :param str main_module_name: Absolute Python-import name of the main module to run.
                                   For example, "example_module".
        :param str module_fs_path: Filesystem path where the module containing the function lives.
                                 For example, "C:\\Users\\eureqa\\example_module".
                                 If omitted, this is inferred by importing the function above and
                                 taking the parent directory of the file that contains it.
        :param list[str] ignore_files: List of names of files that, if encountered, will not be uploaded to the Eureqa server.
        :param list additional_modules_paths: List of other directory names, that, if encountered, will be uploaded to the Eureqa server as well
        :param str icon_path: (Optional) path to an icon to use for the analysis
        """

        with tempfile.TemporaryFile() as f:
            AnalysisTemplate._create_analysis_template_zip(f, main_module_name, module_fs_path, ignore_files, additional_modules_paths)
            self._eureqa._session.report_progress('Setting module for analysis_template: \'%s\' (%s).' % (self.name, self._id))
            raw_data = f.read()
            result = self._eureqa._session.execute('/analysis_templates/%s/script' % (self._id), method='POST', raw_data=raw_data)

        if icon_path is not None:
            with open(icon_path, 'r') as f:
                raw_data = f.read()
                self._eureqa._session.report_progress('Setting icon for analysis_template: \'%s\' (%s).' % (self.name, self._id))
                result = self._eureqa._session.execute('/analysis_templates/%s/icon' % (self._id), method='POST', raw_data=raw_data)

    def get_module(self, output_filename):
        """Download the Python code which implementes this analysis template, as a .zip package,
        to the specified file on the local filesystem. Returns the main_module's name

        :param str output_filename: The filename to be used for the resulting zip file containing the analysis template code.
        """

        self._eureqa._session.report_progress('Getting module for analysis_template: \'%s\' (%s).' % (self.name, self._id))
        contents = self._eureqa._session.execute('/analysis_templates/%s/script' % (self._id), 'GET', raw_return=True)

        with open(output_filename, "wb") as f:
            f.write(contents)

        return AnalysisTemplate._get_main_module_name_from_analysis_template_zip(output_filename)

    def _to_json(self):
        body = {
            'name': self.name,
            'description': self.description,
            'icon_url': self._icon_url,
            'icon_url_fill_custom': self._icon_url_fill_custom,
            'parameter': self.parameters._to_json()
        }
        return body

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    @staticmethod
    def _create_analysis_template_zip(f, main_module_name, module_fs_path, ignore_files, additional_modules_paths=[]):
        """Given arguments described in set_module, packs the files and
        manifest appropriately into f as a .zip package which can be
        executed as an analysis template. Leaves f at the beginning of
        the bytes written.
        """
        with zipfile.ZipFile(f, 'w') as package_zip:
            # Create _entry.py which has a get_main_module_name() funcion
            package_zip.writestr("manifest/eureqa_main_module_name.txt",  main_module_name)
            for module_dir in [ module_fs_path ] + additional_modules_paths:
                AnalysisTemplate._add_module_to_zip(module_dir, package_zip, ignore_files)
        f.seek(0)

    @staticmethod
    def _get_main_module_name_from_analysis_template_zip(file_or_filename):
        with zipfile.ZipFile(file_or_filename, 'r') as package_zip:
            return package_zip.open('manifest/eureqa_main_module_name.txt').read()


    @staticmethod
    def _add_module_to_zip(module_dir, package_zip, _ignore_files = None):

        """Given a module object and a ZipFile object,
        write all files within the module's filesystem directory into the ZipFile."""

        if _ignore_files is None:
            _ignore_files = []

        if not os.path.isdir(module_dir):
            if os.path.isfile(module_dir):
                # Single-file module.  Or maybe just a single file.  Either way, add it.
                package_zip.write(os.path.abspath(module_dir), os.path.basename(module_dir))
                return
            else:
                raise Exception("%s is not an existing file or directory" % (module_dir))

        containing_dir, module_name = os.path.split(module_dir)

        for dirpath, dirnames, filenames in os.walk(module_dir):
            for filename in filenames:
                if filename in _ignore_files:
                    continue
                abs_filename = os.path.join(dirpath, filename)
                assert abs_filename.startswith(containing_dir), \
                  "Internal Error:  File %s should have path starting with %s" % (repr(abs_filename), repr(containing_dir))
                relative_filename = abs_filename[len(containing_dir):]
                package_zip.write(abs_filename, relative_filename)

    def _get_values_for_data_files(self, parameters):
        if parameters is None:
            return
        for p_id in parameters:
            param = parameters[p_id]
            if param._type == "data_file" and param.value is None:
                param._add_file_to_template(self)



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import requests
import json
import time
import warnings

from passwords import _CredentialGetter

from api_version import API_VERSION

def sanitize_name(name):
    #return re.sub(r'[^A-Za-z0-9._]+', '_', name)
    return name

class Http504Exception(Exception):
    pass

class Http404Exception(Exception):
    pass

class _Session:

    def __init__(self, url, user, password, key, verbose, save_credentials, organization, verify_ssl_certificate,
                 retries, timeout_seconds, session_key, interactive_mode):
        """
        Create a new Session object

        Arguments:
            session_key (string): Existing session key

        """
        self.url = url
        self.organization = organization
        self.verbose = verbose
        self.interactive_mode= interactive_mode
        self._session = requests.session()
        self._session.headers["X-nu-api"] = str(API_VERSION)
        self._session.verify = verify_ssl_certificate
        if not verify_ssl_certificate:
            # It is OK to disable warnings because user know what he or she is doing
            # because they explicitly asked to disable SSL verification.
            try:
                requests.packages.urllib3.disable_warnings()
            except Exception:
                warnings.warn("Unable to disable SSL warnings.  Extraneous warnings may be printed.  Try upgrading your 'requests' package.")
        self._retries = retries
        self._timeout_seconds = timeout_seconds
        if session_key is None:
            self._login(user, password, key, save_credentials, organization)
        else:
            self.user = user
            self._set_cookies(user, session_key)
        self._last_response = None

    def _is_root(self): return self.user == "root"

    def execute(self, endpoint, method='POST',args=None, files=None, raw_data=None, raw_returnfile=None, raw_return=False, raw_url=False):
        """Internal function to send a request to the Kerpler instance.

        Arguments:
            endpoint (string): The directory and point to execute on (e.g. '/data1/search3/start')
            method   (string): Type of request, should be 'GET', 'PUT', 'POST', or 'DELETE'
            args     (dict):   The command arguements and values (e.g. { 'start_row':2, 'row_count':10 })
            files    (dict):   Any files to be sent, example: files={'file': open('report.csv', 'rb')}
            raw_data (str):    A JSON string which will be included in a PUT/POST
            raw_return (bool): Return the data as a raw bytestring (str())
            raw_returnfile (str): Optional filepath to save results to
            raw_url (bool):    Treat the URL as-is; don't attempt to prepend '/api' or do other convenience normalization

        Returns:
            The response JSON on success (raises exception otherwise)
        """

        if not endpoint.startswith('/api/v2/') and not raw_url:
            if endpoint.startswith('/api/'):
                api_pos = len('/api/')
                endpoint = endpoint[:api_pos] + 'v2/' + endpoint[api_pos:]
            else:
                endpoint = '/api/v2/%s%s' % (self.organization, endpoint)

        if self.verbose:
            print endpoint

        if self.verbose:
            msg = "\n%4s %s %s" % (method, self.url, endpoint)
            if args:     msg += ' ARGS='     + json.dumps(args, default=str)
            if files:    msg += ' FILES='    + str(files)
            if raw_data: msg += ' RAW_DATA=' + str(raw_data)
            #print(msg[:1000])

        if args and raw_data: raise Exception('Cannot specify both data and raw data')

        url  = self.url + endpoint
        data = None
        if raw_data is not None:
            data = raw_data
        elif args is not None:
            data = json.dumps(args, default=str)

        try:
            if   method == 'GET':    r = self._retry(lambda :self._session.get(url, params=args, timeout=self._timeout_seconds, stream=True))
            elif method == 'PUT':    r = self._retry(lambda :self._session.put(url, data=data, timeout=self._timeout_seconds, stream=True))
            elif method == 'POST':   r = self._retry(lambda :self._session.post(url, data=data, files=files, timeout=self._timeout_seconds, stream=True))
            elif method == 'DELETE': r = self._retry(lambda :self._session.delete(url, params=args, timeout=self._timeout_seconds, stream=True))
            else: raise Exception('Unknown request method: '+method)
        except requests.exceptions.ConnectionError as e:
            print('COULD NOT SEND REQUEST TO SERVER. Check if the web UI is reachable and the server is operational.')
            raise e

        # save response object for debug
        self._last_response = r

        if r.status_code == 504: raise Http504Exception('%s request to URL \'%s\' timed out' % (method, url))
        if r.status_code == 404: raise Http404Exception('URL \'%s\' is not found: \n%s' % (url, str(r.text)))
        if r.status_code != 200: raise Exception('Request unsuccessful (status code %i):\n%s' % (r.status_code, str(r.text)))
        if self.verbose: print('\nRESPONSE ' + str(r.status_code) + ':\n'+r.text[:1000])

        # optionally return raw bytes
        if raw_return:
            r.raw.decode_content = True
            return r.raw.read()

        # optionally save to a file
        if raw_returnfile:
            with open(raw_returnfile, 'wb') as fd:
                for chunk in r.iter_content(20):
                    fd.write(chunk)
            return

        try:
            json_value = json.loads(r.text)
            if isinstance(json_value, dict) and json_value.get('warnings'):
                for w in json_value['warnings']:
                    warnings.warn(w)
            return json_value
        except ValueError:
            return r.text

    def _retry(self, function):
        retries_left = self._retries
        while True:
            try:
                return function()
            except:
                if(retries_left == 0):
                    raise
                time.sleep(1)
                retries_left -= 1

    def _get_user_details(self, user):
        """Get details for a user"""
        self.report_progress('Getting details for user: \'%s\'.' % user)
        result = self.execute('/api/v2/auth/user/' + user, 'GET')
        return result

    def _login(self, user_name=None, password=None, key=None, save_credentials=False, organization=None):
        """Login to a Eureqa instance and store session cookie and organization."""

        if password:
            password = password.encode('base64')

        # lookup or prompt for username/password
        login = _CredentialGetter.get_login(self.url, user_name, password, key)

        try_again = True
        while try_again:
            try:
                self.report_progress('Logging in as user: \'%s\'.' % login.get('username'))
                result = self.execute('/api/v2/auth/login', 'POST', login)
                self._set_cookies(login.get('username'), result['session_key'])
                try_again = False
            except requests.exceptions.ConnectionError as e:
                raise e
            except Exception as e:
                if self.interactive_mode == False:
                    raise e
                elif "two_factor_auth_expected_one_time_password" in e.message:
                    login = _CredentialGetter.prompt_one_time_password(login)
                elif "two_factor_auth_incorrect_one_time_password" in e.message:
                    print e
                    print('\nINCORRECT ONE-TIME PASSWORD. TRY AGAIN OR HIT [ENTER] TO QUIT.\n')
                    login = _CredentialGetter.prompt_one_time_password(login)
                else:
                    print e
                    print('\nLOGIN FAILED. TRY AGAIN OR HIT [ENTER] TO QUIT.\n')
                    login = _CredentialGetter.get_login(self.url, user_name, password, force_prompt=True)

        if save_credentials:
            _CredentialGetter.store_login(self.url, login)

        # get the organization
        user_details_obj = self._get_user_details(login['username'])
        if organization==None:
            self.organization = user_details_obj['organizations'][0] # use first organization
        else:
            self.organization=organization

        self.user = login['username']

    def _set_cookies(self, user_name, session_key):
        """ Set http request cookies for username and session_key """
        cookies = {
            'ncloud_session_key': session_key,
            #Need to pass user name as a cookie, because currently
            #backend misses logic that can infer the current user name
            #from the session data. So it uses value from the cookie.
            'nu_user': user_name.encode('base64')
            #encode('base64') adds trailing new line symbol
            #but the REST end-point does not expect it.
            .rstrip('\n')
        }
        self._session.cookies = requests.utils.cookiejar_from_dict(cookies)


    def _get_session_key(self):
        return self._session.cookies.get('ncloud_session_key')

    def report_progress(self, message, warning=False):
        if warning:
            print >> sys.stderr, "Warning:", message
        elif self.verbose:
            print message

    def report_warning(self, message): self.report_progress(message, warning=True)



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from progress_update import ProgressUpdate
from parameters_values import ParametersValues
from parameter_validation_result import ParameterValidationResult
from eureqa.utils import Throttle

import json

class Execution:
    """Represents an analysis template execution on the server.

    :param dict body: Class metadata as dictionary
    :param str template_id: The id of the analysis_template the execution belongs to.
    :param Eureqa eureqa: A eureqa connection.

    :var str `~eureqa.analysis_templates.execution.Execution.template_id`: The id of the analysis_template the execution belongs to.
    :var str `~eureqa.analysis_templates.execution.Execution.analysis_id`: The id of the analysis the execution belongs to.
    :var str state: The current state of the execution.
    :var list `~eureqa.analysis_templates.execution.Execution.parameters`: The list of parameter values for the execution.
    :var list `~eureqa.analysis_templates.execution.Execution.progress_updates`: The list of updates for the execution.
    """
    def __init__(self, body, template_id, eureqa):
        """For internal use only"""

        self._id = body['id']
        self.template_id = template_id
        self.analysis_id = body['analysis_id']
        self.parameters = ParametersValues._from_json(eureqa, body, execution=self).parameters
        self._add_self_to_data_file_parameters()
        self._eureqa = eureqa
        self._body = body

    @property
    @Throttle()
    def state(self):
        return self._get_updated_self()._body['state']

    def _get_updated_self(self):
        """ Internal. Return a new Execution object representing the current state on the server """
        self._eureqa._session.report_progress('Getting details for execution: \'%s\' of analysis_template: \'%s\'.' % (self._id, self.template_id))
        updated_execution = self._eureqa._session.execute('/analysis_templates/%s/executions/%s' % (self.template_id, self._id), 'GET')
        return Execution(updated_execution, self.template_id, self._eureqa)

    def _is_running(self):
        """Internal (to programatically drive an analysis template). Return
        true if the execution is currently executing or validating"""
        the_state = self.state # NB actually makes a RPC
        return the_state in { 'VALIDATING',  'RUNNING' }

    def _is_done(self):
        """Internal (to programatically drive an analysis template). Return
        true if the execution is currently executing or validating"""
        the_state = self.state # NB actually makes a RPC
        return the_state in { 'DONE', 'ERRORED', 'ABORTED' }

    def _is_waiting_on_confirmation(self):
        """Internal (to programatically drive an analysis template). Return
        true if the execution is waiting on the user to accept results of confirmation"""
        the_state = self.state # NB actually makes a RPC
        return the_state == 'WAITING_FOR_CONFIRMATION'

    @property
    def progress_updates(self):
        """Get all progress updates for an execution of an analysis template."""
        self._eureqa._session.report_progress('Getting progress_updates for execution: \'%s\' of analysis_template: \'%s\'.' % (self._id, self.template_id))
        results = self._eureqa._session.execute('/analysis_templates/%s/executions/%s/progress_updates' % (self.template_id, self._id), 'GET')
        return [ProgressUpdate(x) for x in results]

    @property
    @Throttle()
    def validation_results(self):
        """Get all validation results for the execution of an analysis template."""
        def make_parameter(body):
            r = ParameterValidationResult(type='UNKNOWN')
            r._from_json(body)
            return r
        return [make_parameter(x) for x in self._get_updated_self()._body['validation_results']]

    def _to_json(self):
        body = {
            'template_id': self.template_id,
            'analysis_id': self.analysis_id,
            'state': self._body['state'],
            'parameter': ParametersValues(self.parameters)._to_json(),
            'progress_updates': [x._to_json() for x in self.progress_updates]
        }
        return body

    def _is_running(self):
        """Internal (to programatically drive an analysis template). Return
        true if the execution is currently executing or validating"""
        the_state = self.state # NB actually makes a RPC
        return the_state in { 'VALIDATING', 'RUNNING' }

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    def get_analysis(self):
        """Retrieves the analysis that belongs to the execution."""
        return self._eureqa.get_analysis(self.analysis_id)

    def get_analysis_template(self):
        """Retrieves the analysis that belongs to the execution."""
        return self._eureqa._get_analysis_template(self.template_id)

    def update_progress(self, message):
        """Create a progress update for an execution of an analysis template.

        :param str message: The progress message
        """
        self._eureqa._session.report_progress('Updating progress of execution: \'%s\' of analysis_template: \'%s\'.' % (self._id, self.template_id))
        self._eureqa._session.execute('/analysis_templates/%s/executions/%s/progress_updates' % (self.template_id, self._id), 'POST', { 'message': message })

    def report_validation_result(self, type, message=None, details=None, parameter_id=None):
        """
        Report an info/warning/error message about the specified parameter to be shown to the user in validation review

        :param str type: If this result is INFO, WARNING, or ERROR
        :param str message: The result message
        :param str details: (optional) Detailed message about the result
        :param str parameter_id: (optional) the analysis template parameter that this progress update refers to
        """
        self._eureqa._session.report_progress('Reporting validation result for execution: \'%s\' of analysis_template: \'%s\'.' % (self._id, self.template_id))
        validation_result = ParameterValidationResult(type, message, details, parameter_id)
        self._eureqa._session.execute('/analysis_templates/%s/executions/%s/validation_result' % (self.template_id, self._id), 'POST', validation_result._to_json())

    def report_fatal_error(self, error):
        """Notifies the server that an error occurred during the execution
        and terminates the script execution.
       
        :param str error: The error that occurred during the execution.

        """
        self._eureqa._session.report_progress('Reporting error in execution: \'%s\' of analysis_template: \'%s\'.' % (self._id, self.template_id))
        self._eureqa._session.execute('/analysis_templates/%s/executions/%s/error' % (self.template_id, self._id), 'POST', { 'message': error})
        self._hard_exit()

    def _report_validation_completion(self):
        """Notifies the server that the validation portion has completed."""
        self._eureqa._session.report_progress('Reporting completion of validation: \'%s\' of analysis_template: \'%s\'.' % (self._id, self.template_id))
        self._eureqa._session.execute('/analysis_templates/%s/executions/%s/validation_completion' % (self.template_id, self._id), 'POST', {})

    def _abort(self):
        """Notifies the server that something bad happened to the execution."""
        self._eureqa._session.report_progress('Aborting execution \'%s\' of analysis_template: \'%s\'.' % (self._id, self.template_id))
        self._eureqa._session.execute('/analysis_templates/%s/executions/%s/abort' % (self.template_id, self._id), 'POST', {})

    def _report_validation_results_confirmed(self):
        """Internal (to programatically drive an analysis template). Reports
        that that the user confirmed validation results"""
        self._eureqa._session.report_progress('Reporting validation results were confirmed for execution: \'%s\' of analysis_template: \'%s\'.' % (self._id, self.template_id))
        self._eureqa._session.execute('/analysis_templates/%s/executions/%s/validation_results_confirmed' % (self.template_id, self._id), 'POST', '{}')

    def _report_completion(self):
        """Notifies the server that the execution completed
        and terminates the script execution.
        """
        self._eureqa._session.report_progress('Reporting completion of execution: \'%s\' of analysis_template: \'%s\'.' % (self._id, self.template_id))
        self._eureqa._session.execute('/analysis_templates/%s/executions/%s/complete' % (self.template_id, self._id), 'POST', {})

    def _wait_until_finished(self):
        while not self._is_done():
            if self._is_waiting_on_confirmation():
                self._report_validation_results_confirmed()

    def _hard_exit(self):
        import traceback
        with open('/tmp/we_are_quitting.txt', 'wb') as fd:
            fd.write(traceback.format_exc() + '\n')
        import sys
        sys.exit()

    def _add_self_to_data_file_parameters(self):
        if self.parameters is not None:
            for p_id in self.parameters:
                param = self.parameters[p_id]
                if param._type is "data_file":
                    param._execution = self



"""This module contains components used to build up analyses.

To add a component to an analysis, you use :any:`Analysis.create_card` method::

    h = HtmlBlock('This will render on a Card in an Analysis')
    analysis.create_card(h)

Or the :any:`Analysis.create_html_card` convenience methods::

    vl = VariableLink(datasource=self._data_source, variable_name=self._variables[0])
    analysis.create_html_card("Target variable is: {0}".format(analysis.html_ref(vl)))

"""

from .base import _Component
from .model_fit_by_row_plot import ModelFitByRowPlot
from .model import Model
from .model_fit_separation_plot import ModelFitSeparationPlot
from .model_evaluator import ModelEvaluator
from .distribution_plot import DistributionPlot
from .box_plot import BoxPlot
from .double_histogram_plot import DoubleHistogramPlot
from .scatter_plot import ScatterPlot
from .binned_mean_plot import BinnedMeanPlot
from .by_row_plot import ByRowPlot
from .custom_plot import CustomPlot
from .model_summary import ModelSummary
from .text_block import TextBlock
from .html_block import HtmlBlock
from .titled_layout import TitledLayout
from .magnitude_bar import MagnitudeBar
from .variable_link import VariableLink
from .search_link import SearchLink
from .search_builder_link import SearchBuilderLink
from .download_file import DownloadFile
from .tabbed_layout import TabbedLayout
from .dropdown_layout import DropdownLayout
from .layout import Layout
from .modal import Modal
from .modal_link import ModalLink
from .tooltip import Tooltip
from .table_builder import TableBuilder
from .formatted_text import FormattedText
from .image import Image

__all__ = [
    '_Component',
    'ModelFitByRowPlot',
    'Model',
    'ModelFitSeparationPlot',
    'ModelEvaluator',
    'DistributionPlot',
    'BoxPlot',
    'DoubleHistogramPlot',
    'ScatterPlot',
    'BinnedMeanPlot',
    'ByRowPlot',
    'CustomPlot',
    'ModelSummary',
    'TextBlock',
    'HtmlBlock',
    'TitledLayout',
    'MagnitudeBar',
    'VariableLink',
    'SearchLink',
    'SearchBuilderLink',
    'DownloadFile',
    'TabbedLayout',
    'DropdownLayout',
    'Layout',
    'Modal',
    'ModalLink',
    'Tooltip',
    'TableBuilder',
    'FormattedText',
    'Image'
]



#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from parameter import Parameter

import json

class TextParameter(Parameter):
    """Text parameter description for analysis template

    :param str id: The id of the parameter that will be passed together with its value to an analysis script.
    :param str label: The parameter label that will be shown in UI.
    :param bool text_multiline: Indicates that the text should be split across multiple lines.

    :var str id: The id of the parameter that will be passed together with its value to an analysis script.
    :var str TextParameter.label: The parameter label that will be shown in UI.
    """

    def __init__(self, id, label, text_multiline=False):
        """Initializes a new instance of the ~TextParameter class
        """
        Parameter.__init__(self, id, label, "text_multiline" if text_multiline else "text")

    def __str__(self):
        return json.dumps(self._to_json(), indent=4)

    @staticmethod
    def _from_json(body):
        param = TextParameter(None, None)
        param.from_json(body)
        if param._type not in ["text", "text_multiline"]: raise Exception("Invalid type '%s' specified for text parameter" % param._type)
        return param



# Copyright (c) 2017, Nutonian Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Nutonian Inc nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL NUTONIAN INC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from error_metric import ErrorMetrics

class ModelEvaluation(object):
    """ Represents the result of evaluting solutions and expressions.

    :var dict data: The dictionary, where keys are expressions and variable names, and values are columns of data.
    :var pandas.PandFrame frame: The Pandas frame, where column names are expressions and variable names.
    :var list error_metrics: The list of :py:class:`~eureqa.error_metric.ErrorMetrics` for provided solutions.

    """

    def __init__(self, body):
        """For internal use only
        PARAM_NOT_EXTERNALLY_DOCUMENTED
        """
        self._body = body
        self._columns = body.get('columns')
        self._csv_data = body.get('csv_data')
        self._datetime_column = body.get('datetime_column')
        self._series_id_column = body.get('series_id_column')
        if 'error_metrics' in body:
            error_metrics_array = [ErrorMetrics(_body=value) for value in body.get('error_metrics')]
            self.error_metrics = {value._model: value for value in error_metrics_array}
        else:
            self.error_metrics = None

    @property
    def data(self):
        result = {key: [] for key in self._columns}
        import csv
        import StringIO
        import datetime
        for row in csv.DictReader(StringIO.StringIO(self._csv_data), fieldnames=self._columns):
            for key in self._columns:
                value = row[key]
                if key == self._datetime_column:
                    try:
                        value = datetime.datetime.strptime(row[key],'%m/%d/%Y %H:%M:%S')
                    except ValueError:
                        value = datetime.datetime.strptime(row[key],'%m/%d/%Y')
                elif key == self._series_id_column:
                    pass # Just keep it as string
                else:
                    value = float(value)
                result[key].append(value)
        return result

    @property
    def frame(self):
        from pandas import DataFrame
        return DataFrame(self.data)




#!/usr/bin/env python
#
# This script packages and uploads the python eureqa_api to
# pypi
#
# To publish to pypi (after updating versions) (requires the appropriate account:
#
# python setup.py sdist upload
#
from setuptools import setup, find_packages

setup(name='Eureqa',
      version='1.59.1',
      description='Nutonian Eureqa API',
      author='Nutonian',
      author_email='contact@nutonian.com',
      url='http://nutonian.com/',
      packages=find_packages('.', include=('eureqa','eureqa.*','analysis_templates','analysis_templates.*')),
      install_requires=['requests>=2.7.0'],
      tests_require=['mock>=1.0.1']
     )
